# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Ryan Robison
# Date: 2013jul12

import gpi


class ExternalNode(gpi.NodeAPI):
    """A module for doing least squares fitting of input data to a user
    defined model.  Fitting is always done along the first dimension.

    INPUTS:
    dataIn - data to fit (i.e. "Y" values, assume "X" values are linear)
    coordsIn - "X" values corresponding to "Y" values of dataIn, if they are not linear

    OUTPUTS:
    fit - the fit curve
    coefficients - coefficients of the fit polynomial
    residual - the input data minus the fit curve

    WIDGETS:
    Mode - choose type of fit
    Info - some instructions for use
    For Mode = Polynomial Fit
      Polynomial order - 0 fits to a constant, 1 fits to a line, 2 fits to a parabola, etc.
    For Mode = Generalized Linear 
      data are fit to a linear sum of functions_which_depend_on_x
      # Model Parameters - number of functions to fit (number of coefficients to calculate)
      fi(x) - the model for the ith function, written as a python function of x.  Examples include:
             1                   (i.e. just a constant)
             np.power(x,3)       (fit to x^3)
             np.sin(x)           (fit to sin of x)
    For Mode = Generalized Non-Linear 
      See Info box (with this Mode selected) for detailed information
    """
    def execType(self):
        return gpi.GPI_PROCESS
        #return gpi.GPI_THREAD

    def initUI(self):
        self.nparms = 9
        self.poly_info_text = "Select the order of the polynomial to which " \
                            + "to fit the data.\n"
        self.lin_info_text = "Define each component of the linear model " \
                           + "using the editable\n text boxes below for " \
                           + "each model parameter.\n\n" \
                           + "The linear model follows the relationship:\n"
        self.lin_model = ''
        self.nl_info_text = "Use the Model Definition box below to enter an " \
                          + "expression\nfor the model to which you would " \
                          + "like to fit your data.\n" \
                          + "Follow the format of the example models " \
                          + "shown below.\n\n" \
                          + "    MODEL EXAMPLES:\n" \
                          + "  1) Second order polynomial (3 model " \
                          + "parameters):\n" \
                          + "    y = a0 + a1 * x + a2 * np.power(x, 2)\n" \
                          + "  2) Exponential Decay (2 model parameters):\n" \
                          + "    y = a0 * np.exp(-x/a1)\n" \
                          + "  3) Sinusoid (4 model parameters):\n" \
                          + "    y = a0 + a1 * np.sin(a2*x + a3)\n" \
                          + "  4) Lorentzian (3 model parameters):\n" \
                          + "    y = (a0 / np.pi) * (a1 / 2) / " \
                          + "(np.power(x - a2, 2) + np.power(a1 / 2, 2))\n" \
                          + "  5) Gaussian (3 model parameters):\n" \
                          + "    y = a0 * np.exp(-np.power(x - a1, 2) / " \
                          + "(2 * np.power(a2, 2)))" 


        # Widgets
        self.addWidget('ExclusivePushButtons', 'Mode', buttons=[
                       'Polynomial Fit', 'Generalized Linear',
                       'Generalized Non-linear'], val=0)
        self.addWidget('TextBox', 'Info:', val="")
        self.addWidget('StringBox', 'Model Definition:', val='y = x')
        self.addWidget('SpinBox', '# Model Parameters', val = 3, min=1, max=self.nparms)
        self.addWidget('DoubleSpinBox', 'Initial a0', decimals = 5,  val = 1.0)
        self.addWidget('StringBox', 'f0(x) =', val='1')
        for i in range(self.nparms):
            if i > 0:
                self.addWidget('DoubleSpinBox', 'Initial a%i'%i, decimals = 5,
                               val = 1.0)
                self.addWidget('StringBox', 'f%i(x) ='%i,
                               val='np.power(x, %i)'%i)
        self.addWidget('SpinBox', 'Polynomial Order', val = 3, min=0)
        self.addWidget('PushButton', 'Compute', toggle=True, val=1)

        # IO Ports
        self.addInPort('dataIn', 'NPYarray', obligation=gpi.REQUIRED, ndim = 2)
        self.addInPort('coordsIn', 'NPYarray', obligation=gpi.OPTIONAL, ndim = 1)
        self.addOutPort('fit', 'NPYarray')
        self.addOutPort('coefficients', 'NPYarray')
        self.addOutPort('residual', 'NPYarray')

    def validate(self):

        self.lin_model = 'y = '

        if (self.portEvents() or
            set(self.widgetEvents()).intersection(set(['Mode', '# Model Parameters']))):
            nParms = self.getVal('# Model Parameters')
            mode = self.getVal('Mode')
            for i in range(self.nparms):
                if i < nParms and mode == 1:
                    self.setAttr('f%i(x) ='%i, visible=True)
                    self.lin_model = self.lin_model +'a%i * f%i(x) + '%(i,i)
                else:
                    self.setAttr('f%i(x) ='%i, visible=False)
                if i < nParms and mode == 2:
                    self.setAttr('Initial a%i'%i, visible=True)
                else:
                    self.setAttr('Initial a%i'%i, visible=False)

            if mode == 0:
                self.setAttr('# Model Parameters', visible=False)
                self.setAttr('Model Definition:', visible=False)
                self.setAttr('Polynomial Order', visible=True)
                self.setAttr('Info:', val=self.poly_info_text)
            elif mode == 1:
                self.setAttr('# Model Parameters', visible=True)
                self.setAttr('Model Definition:', visible=False)
                self.setAttr('Polynomial Order', visible=False)
                self.lin_model = self.lin_model[:-3]+'\n'
                self.setAttr('Info:', val=self.lin_info_text
                                  +self.lin_model)
            else:
                self.setAttr('# Model Parameters', visible=True)
                self.setAttr('Model Definition:', visible=True)
                self.setAttr('Polynomial Order', visible=False)
                self.setAttr('Info:', val=self.nl_info_text)

        return(0)

    def compute(self):

        import numpy as np
        from scipy.optimize import curve_fit

        data = self.getData('dataIn')
        coords = self.getData('coordsIn')
        if coords is None:
            coords = np.array(list(range(data.shape[-1])))

        mode = self.getVal('Mode')
        compute = self.getVal('Compute')

        outer_dim = data.shape[0]

        if compute == 1:
            out = np.zeros(data.shape)
            res = np.zeros(data.shape)
            if mode == 0:
                order = self.getVal('Polynomial Order')
                try:
                    p = np.polyfit(coords, data.T, order)
                except:
                    info = self.poly_info_text \
                         + "\n\nError in the data fitting!"
                    self.setAttr('Info:', val=info)
                    self.setAttr('Compute', val=0)
                else:
                    if outer_dim < 20:
                        info = self.poly_info_text \
                            + "\n\nFit coefficients: "+str(p.T)
                        self.setAttr('Info:', val=info)
                    for i in range(outer_dim):
                        a = p[:,i]
                        out[i,:] = np.polyval(a, coords)
                        res[i,:] = np.subtract(data[i,:], out[i,:])
                    self.setData('fit', out)
                    self.setData('coefficients', np.ascontiguousarray(p))
                    self.setData('residual', res)

            elif mode == 1:
                x = coords
                nParms = self.getVal('# Model Parameters')
                A = np.ones([nParms, coords.shape[-1]])
                for i in range(nParms):
                    func = self.getVal('f%i(x) ='%i)
                    exec('A[%i,:] = '%i+func)
                try:
                    a = np.linalg.lstsq(A.T, data.T)[0]
                except:
                    info = self.lin_info_text + self.lin_model \
                         + "\n\nError in the model or data fitting!"
                    self.setAttr('Info:', val=info)
                    self.setAttr('Compute', val=0)
                else:
                    if outer_dim < 20:
                        info = self.lin_info_text + self.lin_model \
                             + "\n\nFit coefficients: "+str(a.T)
                        self.setAttr('Info:', val=info)
                    for i in range(outer_dim):
                        for j in range(nParms):
                            func = self.getVal('f%i(x) ='%j)
                            exec('out[i,:] = out[i,:] + a[j,i] * '+func)
                        res[i,:] = np.subtract(data[i,:], out[i,:])
                    self.setData('fit', out)
                    self.setData('coefficients', np.ascontiguousarray(a))
                    self.setData('residual', res)

            elif mode == 2:
                mod_expr = 'def Model(x, a0, '
                nParms = self.getVal('# Model Parameters')
                regress_code = self.getVal('Model Definition:')
                a0 = []
                for i in range(0, nParms):
                    val = self.getVal('Initial a%i'%i)
                    a0.append(val)
                p0 = a0
                for i in range(1, nParms):
                    mod_expr = mod_expr+'a%i, '%i
                mod_expr = mod_expr[:-2]+'):\n    import numpy as np\n    ' \
                         +regress_code+'\n    return y\n'
                a = np.zeros([outer_dim, nParms])
                try:
                    exec(mod_expr, locals(), globals())
                    for i in range(outer_dim):
                        p, cov = curve_fit(Model, coords, data[i,:], p0=p0)
                        a[i,:] = p
                except:
                    info = self.nl_info_text \
                         + "\n\nError in the model or data fitting at %i!" %i
                    self.setAttr('Info:', val=info)
                    self.setAttr('Compute', val=0)
                else:
                    if outer_dim < 20:
                        info = self.nl_info_text \
                             + "\n\nFit coefficients: "+str(a)
                        self.setAttr('Info:', val=info)
                    for i in range(outer_dim):
                        out_expr = 'out[i,:] = Model(coords, a[i,0], '
                        for j in range(1, nParms):
                            out_expr = out_expr+'a[i,%i], '%j
                        out_expr = out_expr[:-2]+')'
                        exec(out_expr)
                        res[i,:] = np.subtract(data[i,:], out[i,:])
                    self.setData('fit', out)
                    self.setData('coefficients', np.ascontiguousarray(a))
                    self.setData('residual', res)

        return(0)



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Payal Bhavsar
# Date: Dec2013
# built of Ryan Robison's FFT code

import gpi
from gpi import QtWidgets


# WIDGET (code from FFTW_GPI.py)
class FFTW_GROUP(gpi.GenericWidgetGroup):

    """A combination of SpinBoxes, DoubleSpinBoxes, and PushButtons
    to form a unique widget suitable for FFT options on dimensions.
    """
    valueChanged = gpi.Signal()

    def __init__(self, title, parent=None):
        super().__init__(title, parent)

        self._val = {}
        self._val['length'] = 1
        self._val['in_len'] = 1  # the original array length

        self.db = gpi.BasicDoubleSpinBox()  # factor
        self.db.set_label('factor:')
        self.db.set_min(0.001)
        self.db.set_max(gpi.GPI_FLOAT_MAX)
        self.db.set_decimals(3)
        self.db.set_singlestep(0.001)
        self.db.set_val(1)

        self.sb = gpi.BasicSpinBox()  # length
        self.sb.set_label('length:')
        self.sb.set_min(1)
        self.sb.set_val(1)
        self.sb.set_max(gpi.GPI_INT_MAX)

        self.db.valueChanged.connect(self.factChange)
        self.sb.valueChanged.connect(self.lenChange)

        vbox = QtWidgets.QHBoxLayout()
        vbox.addWidget(self.db)
        vbox.addWidget(self.sb)
        vbox.setStretch(0, 0)
        vbox.setStretch(1, 0)
        vbox.setStretch(2, 0)
        vbox.setContentsMargins(0, 0, 0, 0)  # we don't need margins here
        vbox.setSpacing(0)
        self.setLayout(vbox)

    # setters
    def set_val(self, val):
        """A python-dict containing: in_len and length parms. """
        sig = False
        if 'in_len' in val:
            # otherwise this would change every time compute() was called
            if self._val['in_len'] != val['in_len']:
                self._val['in_len'] = val['in_len']
                fact = self.db.get_val()  # set len based on factor
                fact *= self._val['in_len']
                self.setLenQuietly(int(fact))
                self._val['length'] = int(fact)
        if 'length' in val:
            self._val['length'] = val['length']
            self.setLenQuietly(val['length'])
            self.setFactQuietly(
                float(val['length']) / float(self._val['in_len']))
            sig = True
        if sig:
            self.valueChanged.emit()

    # getters
    def get_val(self):
        return self._val

    # support
    def factChange(self, val):
        self._val['length'] = int(self._val['in_len'] * val)
        self.setLenQuietly(self._val['length'])
        self.valueChanged.emit()

    def lenChange(self, val):
        self._val['length'] = val
        self.setFactQuietly(float(val) / float(self._val['in_len']))
        self.valueChanged.emit()

    def setFactQuietly(self, val):
        self.db.blockSignals(True)
        self.db.set_val(val)
        self.db.blockSignals(False)

    def setLenQuietly(self, val):
        self.sb.blockSignals(True)
        self.sb.set_val(val)
        self.sb.blockSignals(False)

class ExternalNode(gpi.NodeAPI):

    """Fast Fourier Transformation of N-dimensional data via scipy library.


    INPUT - data to be zeropadded or sinc-interpolated, can be real or complex.  DC is assumed to be at index N/2 (starting at 0)

    OUTPUT - Return Sinc interpolated data - zero-padding or cropping is applied in the transform domain.

    WIDGETS:
    Dimension i - button turns off/on tranform in ith dimension
                  factor and length are redundant parameters, length is the output dimension size
                  factor = length/(input dimension size)
                  factors < 1 result in cropping data before transformation
                  factors > 1 result in zero-padding before transformation

    compute - return sinc interpolated data after FFT forward the input data -> zero-pad -> FFT backward

    """

    def initUI(self):

        # Widgets
        self.dim_base_name = 'Dimension['
        self.ndim = 9  # limit to 9 for now
        for i in range(self.ndim):
            self.addWidget('FFTW_GROUP', self.dim_base_name +
                           str(-i - 1) + ']')

        self.addWidget('ExclusivePushButtons','Domain', buttons=['Transform','Image'],val=0)
        self.addWidget('PushButton', 'compute', toggle=True)

        # IO Ports
        self.addInPort('in', 'NPYarray', obligation=gpi.REQUIRED)
        self.addOutPort('out', 'NPYarray')

    def validate(self):
        '''update the widget bounds based on the input data
        '''

        # only update bounds if the 'in' port changed.
        if 'in' in self.portEvents():

            data = self.getData('in')

            # visibility and bounds
            for i in range(self.ndim):
                if i < len(data.shape):
                    val = {'in_len': data.shape[-i - 1]}
                    self.setAttr(self.dim_base_name + str(-i - 1) + ']',
                                 visible=True, val=val)
                else:
                    self.setAttr(self.dim_base_name + str(-i - 1) + ']',
                                 visible=False)

        # only change label if the 'direction' widget changed.

        return(0)

    def compute(self):

        import numpy as np

        data = self.getData('in')

        if self.getVal('compute'):

            fftAxes = ()
            ifftAxes = ()

            intype = data.dtype

            temp = data

            for i in range(data.ndim):
                val = self.getVal(self.dim_base_name + str(-i - 1) + ']')

                if val['length'] != val['in_len']:
                    fftAxes = (-i - 1,)
                    ifftAxes = ifftAxes + (-i - 1,)

                    if self.getVal('Domain') == 0:
                        temp = np.fft.fftshift(np.fft.ifftn(np.fft.ifftshift(temp),
                                                        axes=fftAxes))

                    zpad_length = val['length'] - temp.shape[-i - 1]
                    if zpad_length >= 0:
                        zpad_before = int(zpad_length / 2.0 + 0.5)
                        zpad_after = int(zpad_length / 2.0)
                    else:
                        zpad_before = int(zpad_length / 2.0 - 0.5)
                        zpad_after = int(zpad_length / 2.0)
                    if zpad_after > 0:
                        temp = np.insert(temp, temp.shape[-i - 1] *
                                         np.ones(zpad_after), 0.0, (-i - 1))
                    elif zpad_after < 0:
                        temp = np.delete(temp, list(range(temp.shape[-i - 1] +
                                         zpad_after, temp.shape[-i - 1])), (-i - 1))
                    if zpad_before > 0:
                        temp = np.insert(temp, np.zeros(zpad_before), 0.0,
                                         (-i - 1))
                    elif zpad_before < 0:
                        temp = np.delete(temp, list(range(-zpad_before)), (-i - 1))

            if self.getVal('Domain') == 0:
                out = np.fft.fftshift(np.fft.fftn(np.fft.ifftshift(temp),
                                              axes=ifftAxes))
            else:
                out = temp
            self.setData('out', out)

        else:

            self.setData('out', data)

        return 0

    def execType(self):
        '''Could be GPI_THREAD, GPI_PROCESS, GPI_APPLOOP'''
        return gpi.GPI_PROCESS



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


#Author: Daniel Borup
#Date: Jan2019
#Notes: Adapted from Float_Math

import numpy as np
import gpi
import math
from math import exp

class ExternalNode(gpi.NodeAPI):
    """A node to do integer math

    INPUT: input int

    OUTPUT: output float

    WIDGETS:
    For two inputs - math operations - add, subtract, multiply, divide, mod and power
    For single input - math operations - add, subtract, multiply, divide, mod, power by integer and do absolute, exponent, square root

    """

    def initUI(self):

        # Widgets
        self.inputs = 0

        self.addWidget('ExclusiveRadioButtons', 'Operation', buttons=['Add', 'Subtract','Multiply', 'Divide', 'Mod', 'Power','Absolute', 'Exponential', 'Square Root'],val=0)

        self.addWidget('SpinBox', 'Integer', val=1)
        self.addWidget('PushButton', 'compute', toggle=True, val=1)

        # IO Ports
        self.addInPort('inLeft','INT', obligation=gpi.OPTIONAL)
        self.addInPort('inRight','INT', obligation=gpi.OPTIONAL)
        self.addOutPort('out', 'INT')

        # operations

    def validate(self):
        '''update the widgets based on the input arrays
        '''

        data1 = self.getData('inLeft')
        data2 = self.getData('inRight')

        if (set(self.portEvents()).intersection(set(['inLeft','inRight'])) or
            'Mode' in self.widgetEvents()):

            if (data1 is None and data2 is None):
                self.inputs = 0
            elif (data1 is None or data2 is None):
                self.inputs = 1
            else:
                self.inputs = 2

        if self.inputs == 2:
                self.setAttr('Operation', buttons=['Add', 'Subtract','Multiply', 'Divide', 'Mod', 'Power'])

        elif self.inputs == 1:
                self.setAttr('Operation', buttons=['Add', 'Subtract','Multiply', 'Divide', 'Mod', 'Power','Absolute', 'Exponential', 'Square Root'])

        operation = self.getVal('Operation')
        if self.inputs == 1:
            self.setAttr('Integer', visible=True)
        else:
            self.setAttr('Integer', visible=False)

        if operation > 5:
            self.setAttr('Integer', visible=False)

        return(0)


    def compute(self):

        import numpy as np

        data1 = self.getData('inLeft')
        data2 = self.getData('inRight')
        operation = self.getVal('Operation')

        if self.getVal('compute'):
            if self.inputs == 1:
                if data2 is not None:
                    data1 = data2
                data2 = self.getVal('Integer')

            try:
                if self.inputs != 0:
                    if operation == 0:
                        out = data1+data2
                    if operation == 1:
                        out = data1-data2
                    if operation == 2:
                        out = data1*data2
                    if operation == 3:
                        out = data1//data2
                    if operation == 4:
                        out = data1%data2
                    if operation == 5:
                        out = data1**data2
                    if operation == 6:
                        out = abs(data1)
                    if operation == 7:
                        out = math.floor(exp(data1))
                    if operation == 8:
                        out = math.floor(math.sqrt(data1))

                    self.setData('out', out)
            except:
                return 1
        return(0)

    def execType(self):
        '''Could be GPI_THREAD, GPI_PROCESS, GPI_APPLOOP'''
        return gpi.GPI_PROCESS



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Ryan Robison
# Date: 2013jul01

import numpy as np
import gpi


class ExternalNode(gpi.NodeAPI):
    """Unwraps the data of a numpy array containg phase along the specified dimension.

    Widgets:
    Dimension: dimension to unwrap along
    Units:  choose degrees or radians
    """

    def initUI(self):

        # Widgets
        self.addWidget('SpinBox', 'Dimension', val = 0)
        self.addWidget('ExclusivePushButtons', 'Units',buttons=
                        ['degrees','radians','cycles'],val=1)

        # IO Ports
        self.addInPort('in', 'NPYarray', dtype=[np.float64, np.float32],
                       obligation=gpi.REQUIRED)
        self.addOutPort('out', 'NPYarray', dtype=[np.float64, np.float32])

    def validate(self):
        data = self.getData('in')

        self.setAttr('Dimension', max=(len(data.shape)-1),
                          min=(-len(data.shape)))

        return(0)

    def compute(self):

        import numpy as np

        data = self.getData('in')
        dim = self.getVal('Dimension')

        if self.getVal('Units') == 0: #degrees
          out = np.rad2deg(np.unwrap(np.deg2rad(data), discont = np.pi, axis = dim))
        elif self.getVal('Units') == 2: #cycles
          out = np.rad2deg(np.unwrap(np.deg2rad(data*360), discont = np.pi,
                            axis = dim)) / 360
        else: #radians
          out = np.unwrap(data, discont = np.pi, axis = dim)

        self.setData('out', out)
        return(0)

    def execType(self):
        '''Could be GPI_THREAD, GPI_PROCESS, GPI_APPLOOP'''
        return gpi.GPI_PROCESS



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Jim Pipe
# Date: 2013Aug

import gpi

class ExternalNode(gpi.NodeAPI):
    """Node to serially append input data to output data in specified direction

    INPUT: input numpy array

    OUTPUT: data containing glued-together pieces of input data

    WIDGETS:
    Autoadd: when true, any new incoming data automatically gets glued to
        output data
    Add Now: when pressed, existing data at input port is glued to output data
    Glue Dimension: dimension in which to glue input data together
        A negative one here means a "new" dim 0, not the typical python -1
        wrapped dim. Once the gluing starts, this cannot be changed until
        output data are cleared
    Glue Dim Size: Reports back size along the gluing dimension
    Clear Output: Press twice in a row to clear output data (must press
        twice!).
    """

    def initUI(self):
        # Widgets
        self.addWidget('PushButton', 'AutoAdd',
                       button_title='OFF', toggle=True)
        self.addWidget('PushButton', 'Add Now',
                       button_title='ADD', toggle=False)
        self.addWidget('Slider', 'Glue Dimension', min=-1, max=0, val=-1)
        self.addWidget('SpinBox', 'Glue Dim Size', min=0, max=0, val=0)
        self.addWidget('PushButton', 'Clear Output',
                       button_title='RESET OUTPUT', toggle=True)

        # IO Ports
        self.addInPort('indat', 'NPYarray')
        self.addOutPort('outdat', 'NPYarray')

    def validate(self):

        # Must Press 'Clear Output' twice in a row, for safety
        # Any other event after first press will reset 'Clear Output'
        if 'Clear Output' in self.widgetEvents():
            if self.getVal('Clear Output'):
                self.setAttr('Clear Output', button_title='PRESS TO CONFIRM')
            else:
                self.setAttr('Clear Output', button_title='RESET OUTPUT')
        else:
            if self.getVal('Clear Output'):
                self.setAttr('Clear Output', val=0)
                self.setAttr('Clear Output', button_title='RESET OUTPUT')

        # When building data, don't let user change dimensions
        out = self.getData('outdat')
        if out is not None:
            fixdim = self.getVal('Glue Dimension')
            self.setAttr('Glue Dimension', min = fixdim, max = fixdim)
        else:
            data = self.getData('indat')
            if data is not None:
                self.setAttr('Glue Dimension', min = -1, max=data.ndim)

        if self.getVal('AutoAdd'):
            self.setAttr('AutoAdd', button_title='ON')
        else:
            self.setAttr('AutoAdd', button_title='OFF')

        return 0

    def compute(self):
        '''This is where the main algorithm should be implemented.
        '''
        import numpy as np

        indata  = self.getData('indat')
        outdata = self.getData('outdat')
        gd = self.getVal('Glue Dimension')
        gd0 = max(gd, 0) # when gd is -1, need to make it 0 for some calcs

        # Clear output data on 2nd click of 'Clear Output'
        if 'Clear Output' in self.widgetEvents():
            if not self.getVal('Clear Output'):
                outdata = None

        else:
            if self.getVal('AutoAdd') or self.getVal('Add Now'):
                # Add an extra dimension to indata if glueing in that direction
                if gd == indata.ndim:
                    indata = np.expand_dims(indata, gd)
                if gd == -1:
                    indata = np.expand_dims(indata, 0)

                if outdata is None:
                    outdata = indata.copy()
                    self.setAttr('Glue Dimension',min=gd,max=gd,val=gd)
                else:
                    outdata = np.append(outdata, indata,axis=gd0)

        if outdata is not None:
            gds = np.array(outdata.shape)[gd0]
            self.setAttr('Glue Dim Size', min=gds, max=gds, val=gds)
        else:
            self.setAttr('Glue Dim Size', min=0, max=0, val=0)
            if indata is not None:
                self.setAttr('Glue Dimension', min=-1, max=indata.ndim)
            else:
                self.setAttr('Glue Dimension', min=-1, max=0, val=-1)

        self.setData('outdat', outdata)

        return 0

    def execType(self):
        '''Could be GPI_THREAD, GPI_PROCESS, GPI_APPLOOP'''
        return gpi.GPI_THREAD



# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Nick Zwart
# Date: 2012dec10

import gpi


class ExternalNode(gpi.NodeAPI):
    """A basic module for tutorial purposes.  This module contains all
    stock widgets and types so that the node-developer can preview and read
    auto-doc information about each object.
    """

    def initUI(self):
        # Widgets
        self.addWidget('NonExclusivePushButtons', 'My NonExclusivePushButtons',
                       buttons=['hello', 'world', '!!!'], val=[0, 2])
        self.addWidget('TextBox', 'Data Statistics', val='hello data')
        self.addWidget('ComboBox', 'My Exclusive Options', items=['hello', 'world', '!!!'], val='world')
        self.addWidget(
            'PushButton', 'MyPushButton', button_title='OFF', toggle=True)
        self.addWidget(wdg='TextBox', title='MyTextBox', val='No Info Yet...')
        self.addWidget(
            'OpenFileBrowser', 'MyOpenFileBrowser', button_title='Browse',
            caption='Open Stuff', directory='~/',
            filter='my format (*.my)')
        self.addWidget('SaveFileBrowser', 'MySaveFileBrowser')
        self.addWidget('TextEdit', 'MyTextEdit', val="initial text...")
        self.addWidget('ExclusiveRadioButtons', 'MyExclusiveRadioButtons',
                       buttons=['Thread', 'Process', 'Main Loop'],
                       val=2)
        self.addWidget('ExclusivePushButtons', 'MyExclusivePushButtons',
                       buttons=['Thread', 'Process', 'Main Loop'],
                       val=2)
        self.addWidget('DisplayBox', 'MyDisplayBox', collapsed=True)
        self.addWidget('StringBox', 'MyStringBox', val="True")
        self.addWidget('DoubleSpinBox', 'MyDoubleSpinBox', immediate=True)
        self.addWidget('DoubleSpinBox', 'MyDoubleSpinBox2')
        self.addWidget('SpinBox', 'MySpinBox', immediate=True)
        self.addWidget('SpinBox', 'MySpinBox2')
        self.addWidget('Slider', 'MySlider')
        self.addWidget('WebBox', 'myBrowser', val='www.google.com')

        # IO Ports
        self.addInPort('MyNPYarray', 'NPYarray', obligation=gpi.OPTIONAL)
        self.addInPort(title='MyINT', type='INT', obligation=gpi.OPTIONAL)
        self.addInPort(title='MyFLOAT', type='FLOAT', obligation=gpi.OPTIONAL)
        self.addInPort(title='MyLONG', type='LONG', obligation=gpi.OPTIONAL)
        self.addInPort(title='MyLIST', type='LIST', obligation=gpi.OPTIONAL)

        self.addOutPort('MyTUPLE', 'TUPLE')
        self.addOutPort('MyCOMPLEX', 'COMPLEX')
        self.addOutPort('MySTRING', 'STRING')
        self.addOutPort('MyDICT', 'DICT')

        # example of warn logger level.
        self.log.warn("something didnt initialize.")

    def validate(self):
        '''This function runs before compute() as a GPI_APPLOOP exec-type.
        Here, widgets (bounds, limits, etc...) can be modified to ensure they
        are correctly validated before the widget values are used in the
        compute() routine -where widgets are buffered and any modifications are
        applied after compute() runs.
        '''
        self.starttime()  # time your code, NODE level log

        # examples of node and warn logger levels.
        self.log.node("hello from node level logger, running validation()")
        self.log.warn("this is a bad code area")

        # the path to the gpi.py main script
        self.log.node("gpi.py path:"+gpi.defines.GPI_CWD)

        print("template validation")
        print("DoubleSpinBox: ", self.getVal('MyDoubleSpinBox'))
        print("DoubleSpinBox: ", self.getVal('MyDoubleSpinBox2'))
        print("SpinBox: ", self.getVal('MySpinBox'))
        print("SpinBox2: ", self.getVal('MySpinBox2'))

        # check for misc event types, these can also be used in compute
        print(self.getEvents()) # super set of events
        print(self.portEvents())
        print(self.widgetEvents())

        # validate widget bounds
        data = self.getData('MyNPYarray')
        if data:
            self.setAttr('MyDoubleSpinBox', max=data.shape[0])

        self.endtime('validate()')  # endtime w/ message
        self.endtime()  # endtime w/o message

        return 0

    def compute(self):
        '''This is where the main algorithm should be implemented.
        '''

        self.log.node("hello from node level logger, running compute()")

        print("template compute")

        # GETTING WIDGET INFO
        val = self.getVal('MyPushButton')
        button_title = self.getAttr('MyPushButton', 'button_title')

        # SETTING WIDGET INFO
        if val:
            self.setAttr('MyPushButton', button_title="ON")
        else:
            self.setAttr('MyPushButton', button_title="OFF")

        self.setAttr('MyTextBox', val="The button title was: "+button_title)

        # GETTING PORT INFO
        data = self.getData('MyNPYarray')

        # SETTING PORT INFO
        self.setData('MyTUPLE', ('hello', 'world', 42))
        self.setData('MyCOMPLEX', 3 + 0.14159j)

        # if setData() isn't called, then no downsteam events will trigger
        # and the port data will remain unchanged.
        if self.getData('MySTRING') is None:
            self.log.node('set string')
            self.setData('MySTRING', '/home/bni/data/propellor/lunchtime/meatloaf')
        else:
            pass

        # if the outport is set to None, then no downstream events will trigger
        out = {'hello':1}
        if not self.getData('MyDICT'):
            self.setData('MyDICT', out)
        else:
            self.setData('MyDICT', None)

        return 0

    def execType(self):
        '''Could be GPI_THREAD, GPI_PROCESS, GPI_APPLOOP'''
        return gpi.GPI_PROCESS



# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


#Author: Sudarshan Ragunathan
#Date: 2013aug05

import gpi
import sys
import traceback
import numpy as np
from math import fabs, sqrt, exp

class ExternalNode(gpi.NodeAPI):
    """Collapse data along selected dimension

        INPUT - input array

        OUTPUTS:
        Collapse Single Dim - output array, with 1 less dimension than input array (collapsed version)
        Collapse All Dims - single float value from taking operation on entire array, active when "Collapse All" is selected
        Max Val Index - returns the index of the max value when used with collapse all (type : list)

        WIDGETS:
        Status,Info - information boxes
        Operation - selected method of collapse
        Min,Max,Mean,Std. Dev,Sum,Prod,Median - self-evident
        RMS - performs rms over the magnitude of complex input, else performs rms directly on input
        Energy - sum of squares along dimension
        SWA - Energy/Sum
        Max Val Index - index along dimension at which the maximum value occurs
        Geo-Avg - Nth root of Prod, where N is the size of the collapse dimension
        Dimension - dimension along with to collapse
        Compute - compute
        Span Entire Dimension - select whether to collapse along the entire span of specified dimension
        Dimension Start_Index - if Span Entire Dimension is off, lets you pick index of where collapse starts
        Dimension Stop_Index  - if Span Entire Dimension is off, lets you pick index of where collapse ends
        Collapse_All - when off, module collapses along single (specified) dimension, output array at Collapse Dim
                   when on, module collapses entire data set, output value at Collapse All
        Non-Zero - Used with Collapse All to perform collapse on non-zero values of ndarray
    """

    def initUI(self):

        # Widgets
        self.addWidget('TextBox', 'Status', val='Ready.')
        self.addWidget('TextBox', 'Info', val='Displays single value float results only')
        self.maxdim = 13
        self.ndim = self.maxdim
        self.dim = 0
        self.size = 0
        self.op_buttons = ['Min','Max','Mean','Std. Dev','Sum','RMS','Energy','SWA','Max Val Index','Prod','Geo-Avg','Median']
        self.addWidget('ExclusiveRadioButtons','Operation', buttons=self.op_buttons, val=0)
        dim_buttons = []
        for i in range(-1, -5, -1):
           dim_buttons.append(str(i))
    
        self.addWidget('PushButton', 'Compute', toggle=True, val=1)
        self.addWidget('PushButton', 'Collapse All', toggle=True, val=0)
        self.addWidget('ExclusivePushButtons','Dimension',buttons=dim_buttons, val=0)    
        self.addWidget('PushButton', 'Span Entire Dimension', toggle=True, val=1)
        self.addWidget('Slider', 'Dimension Start_Index', min=0, max=1, val=0)
        self.addWidget('Slider', 'Dimension Stop_Index', min=0, max=1, val=1)
        self.addWidget('PushButton', 'Non-Zero', toggle=True, val=0)
        
            # IO Ports
        self.addInPort('in', 'NPYarray', obligation=gpi.REQUIRED)
        self.addOutPort('Collapse Single Dim','NPYarray')
        self.addOutPort('Collapse All Dims','FLOAT')
        self.addOutPort('Max Val Index','PASS')
    
    def validate(self):

        data = self.getData('in')
        dim_number = self.getVal('Dimension')
        self.dim = int(-(dim_number+1))
        try:
          dlen = data.shape[self.dim]
        except (ValueError, IndexError):
          self.log.warn("Chosen Dimension does not exist. Please make another selection")
          return 1
        dim_buttons = []
        for i in range(-1, -(data.ndim+1), -1):
            dim_buttons.append(str(i))
        self.setAttr('Dimension', buttons=dim_buttons)    
    
        # Change visibility of Start and Stop Index Sliders 
        if self.getVal('Collapse All'):
          self.setAttr('Dimension', visible=False)
          self.setAttr('Span Entire Dimension', visible=False)
          self.setAttr('Dimension Start_Index', visible=False)
          self.setAttr('Dimension Stop_Index', visible=False)
        else:
          self.setAttr('Dimension', visible=True)
          self.setAttr('Span Entire Dimension', visible=True)
          if self.getVal('Span Entire Dimension'):
            self.setAttr('Dimension Start_Index', visible=False,min=0,max=0)
            self.setAttr('Dimension Stop_Index', visible=False,min=0,max=0)
          else:
            self.setAttr('Dimension Start_Index',visible=True,min=0,max=dlen)
            self.setAttr('Dimension Stop_Index',visible=True,min=0,max=dlen)
    
        # Set Collapse all if # dimensions = 1
        if data.ndim == 1:
            self.setAttr('Collapse All', val=1)
            print(str(self.dim)) 

        # Check for Start Index to never exceed Stop Index
        I_maxval = self.getVal('Dimension Start_Index')
        F_maxval = self.getVal('Dimension Stop_Index')
        if I_maxval > F_maxval:
            self.setAttr('Dimension Start_Index', val=np.maximum(I_maxval,F_maxval))
            self.setAttr('Dimension Stop_Index', val=np.maximum(I_maxval,F_maxval))    

        # Check for Compute enabled/disabled
        if self.getVal('Compute')==0:
            self.setData('Collapse Single Dim',None)
            self.setData('Collapse All Dims',None)
            self.setData('Max Val Index',None)
            
        self.setAttr('Non-Zero',visible=True)
    

    def compute(self):

        import sys
        import numpy as np
        import scipy as sp
        data_in = np.dtype(float)
        data_in = self.getData('in')
        op = self.getVal('Operation')
        nonzero = self.getVal('Non-Zero')
    
        dim_number = self.getVal('Dimension')
        self.dim = int(-(dim_number+1))
        try:
          dlen = data_in.shape[self.dim]
        except (ValueError, IndexError):
          self.log.warn("Chosen Dimension does not exist. Please make another selection")
          return 1
        temp1_data = data_in.swapaxes(self.dim,0)
        self.setAttr('Dimension Start_Index', min=0, max=dlen)
        self.setAttr('Dimension Stop_Index', min=0, max=dlen)
        if self.getVal('Span Entire Dimension'):
            temp2_data = temp1_data
        else:
            I_maxval = self.getVal('Dimension Start_Index')
            F_maxval = self.getVal('Dimension Stop_Index')
            temp2_data = temp1_data[I_maxval:F_maxval+1]
        data = temp2_data.swapaxes(self.dim,0)
        if self.getVal('Compute'):
            if op == 0:    # Min
                if self.getVal('Collapse All')==0:
                    if nonzero:
                        tempData = np.zeros_like(data)
                        np.copyto(tempData,data)
                        np.copyto(tempData,np.nan,where=(tempData == 0))
                        collapse_dim = np.nanmin(tempData, axis=self.dim)
                        np.nan_to_num(collapse_dim,copy = False)
                    else:
                        collapse_dim = np.amin(data, axis=self.dim)
                else:
                    if nonzero:
                        collapse_all = np.amin(data[np.nonzero(data)])
                    else:
                        collapse_all = np.amin(data)
            if op == 1:    # Max
                if self.getVal('Collapse All')==0:
                    if nonzero:
                        tempData = np.zeros_like(data)
                        np.copyto(tempData,data)
                        np.copyto(tempData,np.nan,where=(tempData == 0))
                        collapse_dim = np.nanmax(tempData, axis=self.dim)
                        np.nan_to_num(collapse_dim,copy = False)
                    else:
                        collapse_dim = np.amax(data, axis=self.dim)
                else:
                    if nonzero:
                        collapse_all = np.amax(data[np.nonzero(data)])
                    else:
                        collapse_all = np.amax(data)
            if op == 2:    # Mean
                if self.getVal('Collapse All')==0:
                    if nonzero:
                        tempData = np.zeros_like(data)
                        np.copyto(tempData,data)
                        np.copyto(tempData,np.nan,where=(tempData == 0))
                        collapse_dim = np.nanmean(tempData, axis=self.dim)
                        np.nan_to_num(collapse_dim,copy = False)
                    else:
                        collapse_dim = np.mean(data, axis=self.dim)
                else: # collapse all
                    if nonzero:
                        collapse_all = np.mean(data[np.nonzero(data)])
                    else:
                        collapse_all = np.mean(data)
            if op == 3:    # Standard Deviation
                if self.getVal('Collapse All')==0:
                    if nonzero:
                        tempData = np.zeros_like(data)
                        np.copyto(tempData,data)
                        np.copyto(tempData,np.nan,where=(tempData == 0))
                        collapse_dim = np.nanstd(tempData, axis=self.dim)
                        np.nan_to_num(collapse_dim,copy = False)
                    else:
                        collapse_dim = np.std(data, axis=self.dim)
                else:
                    if nonzero:
                        collapse_all = np.std(data[np.nonzero(data)])
                    else:
                        collapse_all = np.std(data)
            if op == 4:    # Sum
                if self.getVal('Collapse All')==0:
                    collapse_dim = np.sum(data, axis=self.dim)
                else:
                    if nonzero:
                        collapse_all = np.sum(data[np.nonzero(data)])
                    else:
                        collapse_all = np.sum(data)
            if op == 5:    # RMS
                data_type = str(data.dtype)
                if ('float' in data_type) or ('int' in data_type):
                    temp = data
                    temp_nz = data[np.nonzero(data)]
                elif 'complex' in data_type:
                    data_mag = np.abs(data)
                    temp = data_mag
                    temp_nz = data_mag[np.nonzero(data_mag)]
                if self.getVal('Collapse All')==0:
                    if nonzero:
                        tempData = np.ones_like(temp)
                        np.copyto(tempData,temp)
                        np.copyto(tempData,np.nan,where=(tempData == 0))
                        temp1_sq = np.square(tempData)
                        temp1_msq = np.nanmean(temp1_sq, axis=self.dim)
                        np.nan_to_num(temp1_msq,copy = False)
                        collapse_dim = np.sqrt(temp1_msq)
                    else:
                        temp1_sq = np.square(temp)
                        temp1_msq = np.mean(temp1_sq, axis=self.dim)
                        collapse_dim = np.sqrt(temp1_msq)
                else:
                    if nonzero:
                        temp2_sq = np.square(temp_nz)
                        temp2_msq = np.mean(temp2_sq)
                        collapse_all = np.sqrt(temp2_msq)
                    else:
                        temp2_sq = np.square(temp)
                        temp2_msq = np.mean(temp2_sq)
                        collapse_all = np.sqrt(temp2_msq)
            if op == 6:    # Energy
                temp_sq = np.square(data)
                tempnz_sq = np.square(data[np.nonzero(data)])
                if self.getVal('Collapse All')==0:
                    collapse_dim = np.sum(temp_sq, axis=self.dim)
                else:
                    if nonzero:
                        collapse_all = np.sum(tempnz_sq)
                    else:
                        collapse_all = np.sum(temp_sq)
            if op == 7:    # Self Weighted Avg.
                if self.getVal('Collapse All')==0:
                    temp1_sos = np.sum(np.square(data), axis=self.dim)
                    temp1_sum = np.sum(data, axis=self.dim)
                    collapse_dim = np.divide(temp1_sos, temp1_sum)
                else:
                    if nonzero:
                        temp2_sos = np.sum(np.square(data[np.nonzero(data)]))
                        temp2_sum = np.sum(data[np.nonzero(data)])
                        collapse_all = np.divide(temp2_sos, temp2_sum)
                    else:
                        temp2_sos = np.sum(np.square(data))
                        temp2_sum = np.sum(data)
                        collapse_all = np.divide(temp2_sos, temp2_sum)
            if op == 8:    # Max Value Index
                if self.getVal('Collapse All')==0:
                    if nonzero:
                        stripMin = np.amin(data, axis = self.dim, keepdims = True)
                        tempData = np.zeros_like(data)
                        np.copyto(tempData,data)
                        tempData = tempData - stripMin
                        np.copyto(tempData,0,where = (data == 0))
                        collapse_dim = np.argmax(tempData, axis=self.dim)
                        np.nan_to_num(collapse_dim,copy = False)
                    else:
                        collapse_dim = np.argmax(data, axis=self.dim)
                else:
                    if nonzero:
                        globalMin = np.amin(data)
                        tempData = np.zeros_like(data)
                        np.copyto(tempData,data)
                        tempData = data - globalMin
                        np.copyto(tempData,0,where = (data == 0))
                        collapse_all = np.unravel_index(tempData.argmax(),data.shape)
                    else:
                        collapse_all = np.unravel_index(data.argmax(),data.shape)
            if op == 9:    # Product
                if self.getVal('Collapse All')==0:
                    if nonzero:
                        tempData = np.zeros_like(data)
                        np.copyto(tempData,data)
                        np.copyto(tempData,np.nan,where=(tempData == 0))
                        collapse_dim = np.nanprod(tempData, axis=self.dim)
                    else:
                        collapse_dim = np.prod(data, axis=self.dim)
                else:
                    if nonzero:
                        collapse_all = np.prod(data[np.nonzero(data)])
                    else:
                        collapse_all = np.prod(data)
            if op == 10:    # Geometric Avg.
                if self.getVal('Collapse All')==0:
                    if nonzero:
                        tempData = np.zeros_like(data)
                        np.copyto(tempData,data)
                        temp1_size = np.count_nonzero(tempData, axis=self.dim).astype(float)
                        np.copyto(tempData,np.nan,where=(tempData == 0))
                        temp1_prod = np.nanprod(tempData, axis=self.dim)
                        # Note - the two lines below would give a complex geo-mean for negative products
                        # But I will leave this out for now so it just throws a warning and gives 0
                        # if(np.amin(temp1_prod) < 0):
                        #     temp1_prod = temp1_prod.astype(complex)
                        temp1_exp = np.divide(1.,temp1_size, out = np.zeros_like(temp1_size), where = (temp1_size != 0))
                        collapse_dim = np.power(temp1_prod, temp1_exp, out = np.zeros_like(temp1_prod),where = (temp1_size != 0))
                        np.nan_to_num(collapse_dim,copy = False)
                    else:
                        temp1_prod = np.prod(data, axis=self.dim)
                        temp1_size = np.size(data, axis=self.dim)
                        collapse_dim = np.power(temp1_prod, (1/temp1_size))
                else:
                    if nonzero:
                        temp2_prod = np.prod(data[np.nonzero(data)])
                        temp2_size = np.size(data[np.nonzero(data)])
                        collapse_all = np.power(temp2_prod, (1/temp2_size))
                    else:
                        temp2_prod = np.prod(data)
                        temp2_size = np.size(data)
                        collapse_all = np.power(temp2_prod, (1/temp2_size))
            if op == 11:    # Median
                if self.getVal('Collapse All')==0:
                    if nonzero:
                        tempData = np.zeros_like(data)
                        np.copyto(tempData,data)
                        np.copyto(tempData,np.nan,where=(tempData == 0))
                        collapse_dim = np.nanmedian(tempData, axis=self.dim)
                        np.nan_to_num(collapse_dim,copy = False)
                    else:
                        collapse_dim = np.median(data, axis=self.dim)
                else:
                    if nonzero:
                        collapse_all = np.median(data[np.nonzero(data)])
                    else:
                        collapse_all = np.median(data)


            if self.getVal('Collapse All'):
                if op == 8:
                    maxval_dict = {'index':collapse_all}
                    self.setData('Max Val Index', list(collapse_all))
                    temp_index = np.asarray(collapse_all)
                    self.setData('Collapse All Dims', float(np.prod(temp_index)))
                else:
                    self.setData('Collapse All Dims', float(collapse_all))
                    self.setData('Max Val Index', None)
                self.setData('Collapse Single Dim', None)
                self.setAttr('Status', val='Ready')
                out_op = self.op_buttons[op]
                info = out_op+" = "+str(collapse_all)+"\n"
                self.setAttr('Info', val=info)
            else:
                self.setData('Collapse Single Dim', collapse_dim)
                self.setData('Collapse All Dims', None)
                self.setData('Max Val Index', None)
                self.setAttr('Status', val='Ready')
                info = "input: "+str(data.shape) +"\noutput: "+str(collapse_dim.shape)
                self.setAttr('Info', val=info)

        return(0)


    def execType(self):
#        '''Could be GPI_THREAD, GPI_PROCESS, GPI_APPLOOP'''
         return gpi.GPI_PROCESS



# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Nick Zwart
# Date: 2012dec10

import gpi


class ExternalNode(gpi.NodeAPI):
    """A node that allows cyclic connections and provides a countdown for
    iterative algorithms.

    INPUTS:
    n+1 - input for iterations, typically data taken from output "n" and processed 
    n_0 - initial condition

    OUTPUT:
    n - value of data at nth iteration

    WIDGETS:
    Current Iteration - reports current iteration index
    Max Iteration - enter number of desired iterations
    Start/Stop - starts/stops iterations
    Reset - resets iteration count to zero and data to initial condition (at input n_0)
    """

    def initUI(self):
        # Widgets
        self.addWidget(wdg='TextBox', title='Current Iteration', val='0')
        self.addWidget(wdg='SpinBox', title='Max Iteration', min=0)
        self.addWidget(wdg='PushButton', title='Start/Stop', toggle=True)
        self.addWidget(wdg='PushButton', title='Reset')

        # IO Ports
        self.addInPort(title='n+1', type='PASS', obligation=gpi.OPTIONAL, cyclic=True)
        self.addInPort(title='n_0', type='PASS', obligation=gpi.REQUIRED)
        self.addOutPort(title='n', type='PASS')

    def validate(self):
        '''This function runs before compute() as a GPI_APPLOOP exec-type.
        Here, widgets (bounds, limits, etc...) can be modified to ensure they
        are correctly validated before the widget values are used in the
        compute() routine -where widgets are buffered and any modifications are
        applied after compute() runs.
        '''

        # if init changes then reset the count and stop iteration
        if 'n_0' in self.portEvents():
            self.setAttr('Reset', val=True)

        return 0

    def compute(self):
        '''This is where the main algorithm should be implemented.
        '''
        reset = self.getVal('Reset')
        on = self.getVal('Start/Stop')
        iternum = int(self.getVal('Current Iteration'))
        maxiter = self.getVal('Max Iteration')
        np1 = self.getData('n+1')
        init = self.getData('n_0')

        if reset:
            self.setAttr('Start/Stop', val=False)
            self.setAttr('Current Iteration', val='0')
            self.setData('n', init)

        elif on:
            if iternum >= maxiter:
                self.setAttr('Start/Stop', val=False)
                return 0
            elif iternum == 0:
                self.setAttr('Current Iteration', val=str(iternum+1))
                self.setData('n', init)
            else:
                self.setAttr('Current Iteration', val=str(iternum+1))
                self.setData('n', np1)

        else:
            self.setData('n', None)

        if (maxiter-1) > 0:
            self.setStatus(str(iternum*100.0/(maxiter-1)) + "%")

        return 0

    def execType(self):
        '''Could be GPI_THREAD, GPI_PROCESS, GPI_APPLOOP'''
        return gpi.GPI_THREAD



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Jim Pipe
# Date: 2013 Oct

import gpi
from gpi import QtGui
import numpy as np

class ExternalNode(gpi.NodeAPI):
    """2D image Compare Module

    INPUTS (must be the same size):
    3D uint8 ARGB data (e.g. from ImageDisplay)
    3D uint8 ARGB data (e.g. from ImageDisplay)

    OUTPUT:
    3D data of displayed image, last dimension has length 4 for ARGB byte (uint8) data

    WIDGETS:
    Transition: chooses how to transition between images of left and right ports
    LeftRight: Toggles between left or right images
    edge: slider to demarcate the line, or fading, between two images
    """

    def execType(self):
        return gpi.GPI_APPLOOP

    def initUI(self):

        # Widgets
        self.addWidget('DisplayBox', 'Viewport:')
        self.addWidget('ExclusivePushButtons','Transition',
                       buttons=['Toggle','Fade','Hor','Vert','Color'], val=0)
        self.addWidget('PushButton', 'LeftRight', button_title='Left Port', toggle=True)
        self.addWidget('Slider', 'edge',val=0)

        # IO Ports
        self.addInPort('inleft', 'NPYarray', ndim=3, obligation=gpi.REQUIRED)
        self.addInPort('inright', 'NPYarray', ndim=3, obligation=gpi.REQUIRED)
        self.addOutPort('out', 'NPYarray')

    def validate(self):

        # Complex or Scalar?
        inleft = self.getData('inleft')
        inright = self.getData('inright')

        if (inleft.shape[-1] != 4):
            self.log.warn("left port data must have a last dimension of 4")
            return 1
        if (inright.shape[-1] != 4):
            self.log.warn("right port data must have a last dimension of 4")
            return 1
        if (inleft.shape != inright.shape):
            self.log.warn("data must be the same size")
            return 1

        if self.getVal('Transition') == 0: # Toggle
          self.setAttr('edge',visible=False)
          if self.getVal('LeftRight'):
            self.setAttr('LeftRight',button_title="Right Port")
          else:
            self.setAttr('LeftRight',button_title="Left Port")
        elif self.getVal('Transition') == 1: # Fade
          self.setAttr('edge',visible=True,max=100)
          if self.getVal('LeftRight'):
            self.setAttr('LeftRight',button_title="Right Port at edge=0")
          else:
            self.setAttr('LeftRight',button_title="Left Port at edge=0")
        elif self.getVal('Transition') == 2: # Horizontal
          self.setAttr('edge',visible=True,max=inleft.shape[0])
          if self.getVal('LeftRight'):
            self.setAttr('LeftRight',button_title="Right Port on top")
          else:
            self.setAttr('LeftRight',button_title="Left Port on top")
        elif self.getVal('Transition') == 3: # Vertical
          self.setAttr('edge',visible=True,max=inleft.shape[1])
          if self.getVal('LeftRight'):
            self.setAttr('LeftRight',button_title="Right Port on left")
          else:
            self.setAttr('LeftRight',button_title="Left Port on left")
        elif self.getVal('Transition') == 4: # Color
          self.setAttr('edge',visible=True,max=5)
          if self.getVal('LeftRight'):
            self.setAttr('LeftRight',button_title="Left Port RYGCBM")
          else:
            self.setAttr('LeftRight',button_title="Right Port RYGCBM")

        return 0

    def compute(self):

        edgeval = self.getVal('edge')

        # make a copy for changes
        if self.getVal('LeftRight'):
          outright = self.getData('inleft')
          outleft  = self.getData('inright')
        else:
          outleft = self.getData('inleft')
          outright = self.getData('inright')

        if self.getVal('Transition') == 0: # Toggle
          out = outleft
        elif self.getVal('Transition') == 1: # Fade
          cr = 0.01*float(self.getVal('edge'))
          cl = 1.-cr
          out = cl*outleft.astype(np.float) + cr*outright.astype(np.float)
        elif self.getVal('Transition') == 2: # Horizontal
          out = np.append(outleft[:edgeval,:,:],outright[edgeval:,:,:],axis=0)
        elif self.getVal('Transition') == 3: # Vertical
          out = np.append(outleft[:,:edgeval,:],outright[:,edgeval:,:],axis=1)
        elif self.getVal('Transition') == 4: # Color
          out = np.copy(outleft)
          if self.getVal('edge') <= 1 or self.getVal('edge') == 5:
            out[:,:,2] = outright[:,:,2] # Red
          if self.getVal('edge') >= 1 and self.getVal('edge') <= 3:
            out[:,:,1] = outright[:,:,1] # Green
          if self.getVal('edge') >= 3:
            out[:,:,0] = outright[:,:,0] # Blue


        image1 = out.astype(np.uint8)

        h, w = out.shape[:2]
        format_ = QtGui.QImage.Format_RGB32

        image = QtGui.QImage(image1.data, w, h, format_)
        image.ndarray = image1
        if image.isNull():
            self.log.warn("Image Viewer: cannot load image")

        self.setAttr('Viewport:', val=image)
        self.setData('out',image1)

        return 0



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


#!/usr/bin/env python
#__author__ = "Mike Schar"
#__date__ = "2013dmar27"

import gpi
from gpi import QtWidgets

# WIDGET
class ComboBox_GROUP(gpi.GenericWidgetGroup):
    """A combination of two comboBoxes for image quality measure of two images next to each other.
        """
    valueChanged = gpi.Signal()

    def __init__(self, title, parent=None):
        super().__init__(title, parent)

        self._val = {}
        self._val['image_quality_left'] = 0
        self._val['image_quality_right'] = 0

        #image_quality_left
        self.iql = gpi.ComboBox('image_quality_left')
        self.iql.set_items(['choose..', 'excellent', 'good', 'diagnostic', 'non-diagnostic'])
        #self.iql.set_val('choose..')
        self.iql.set_index(0)

        #image_quality_right
        self.iqr = gpi.ComboBox('image_quality_right')
        self.iqr.set_items(['choose..', 'excellent', 'good', 'diagnostic', 'non-diagnostic'])
        #self.iqr.set_val('choose..')
        self.iqr.set_index(0)

        self.iql.valueChanged.connect(self.iqlChange)
        self.iqr.valueChanged.connect(self.iqrChange)

        vbox = QtWidgets.QHBoxLayout()
        vbox.addWidget(self.iql)
        vbox.addWidget(self.iqr)
        vbox.setStretch(0, 0)
        vbox.setStretch(1, 0)
        vbox.setContentsMargins(0, 0, 0, 0)  # we don't need margins here
        vbox.setSpacing(0)
        self.setLayout(vbox)

    # setters
    def set_val(self, val):
      """A python-dict containing: image_quality_left and image_quality_right parms. """
      if 'image_quality_left' in val:
        # check if identical so this does not change every time compute is called (according to code in FFTW)
        if self._val['image_quality_left'] != val['image_quality_left']:
          self._val['image_quality_left'] = val['image_quality_left']
          self.iql.blockSignals(True)
          self.iql.set_index(val['image_quality_left'])
          self.iql.blockSignals(False)

      if 'image_quality_right' in val:
        if self._val['image_quality_right'] != val['image_quality_right']:
          self._val['image_quality_right'] = val['image_quality_right']
          self.iqr.blockSignals(True)
          self.iqr.set_index(val['image_quality_right'])
          self.iqr.blockSignals(False)

    # getters
    def get_val(self):
      return self._val

    # support
    def iqlChange(self, val):
      self._val['image_quality_left'] = val
      self.valueChanged.emit()

    def iqrChange(self, val):
      self._val['image_quality_right'] = val
      self.valueChanged.emit()



def toggle_compare(x):
  if x == 0:
    return 5
  elif x == 1:
    return 4
  elif x== 2:
    return 2
  elif x == 3:
    return 3
  elif x == 4:
    return 1
  elif x == 5:
    return 0
  else:
    return -1

class ExternalNode(gpi.NodeAPI):
    """ Module to compare images in a blinded and randomized manner.
        Input Ports:
        Input_List port requires a list of numpy arrays.
           Each numpy array should have the following dimensions:
              [different reconstructions to be compared (2 or 3); number of slices; x resolution; y resolution]
           To combine different numpy arrays to a list copy the following code into a custom module:
              in1 = self.getData('in1')
              in2 = self.getData('in2')
              # if in1 is a list, append in2 to the list, if in1 is not a list, combine them to a list
              if type(in1) is list:
                in1.append(in2)
                out = in1
              else:
                out = [in1, in2]
              self.setData_ofPort('out1', out)
        previous_analysis_array port is an optional port to load a previously stored analysis array (numpy array)
           to continue the interrupted work.

        Widgets:
        image comparison: select whether the image above the button is much better, better, or the same as the other image.
           Use much better for obviously better images.
           Use better if you require the Toggle button to detect small differences.
           Use same if the quality is the same even after using the Toggle button.
           If image quality rating is disabled, then the next slice will be displaced after pressing a button.
        image quality: select the image quality of the image above as either excellent, good, diagnostic, or non-diagnostic.
           This widget may not be visible in case the "Enable image quality rating" button is set to "Off".
        Toggle: Toggle button to switch the image on the left with the image on the right.
        current slice: Shows how many slices have been analyzed so far.
        out of slices: Shows the number of slices to analyze.
        Undo button: Press this button to undo the last selection made.
        Enable image quality rating: The image quality rating widget is only visible if this button is set to On.
           This button is only visible if the current slice is 0 (zero).

        Output Ports:
        image_for_display port is a 2-dimensional numpy array that should be connected to an ImageDisplay module.
        analysis_array: The numpy array with the results of the analsys.
           Use SaveNPY to store the results. Make sure to store every so often as a backup of your work.
              You can load the stored backup by using LoadNPY and connecting to previous_analysis_array input port.
           Data are stored as follows:
              [nr_slices, 4/7]
              per slice up to 4/7 (number of reconstructions 2/3)4/7 (number of reconstructions 2/3)4/7 (number of reconstructions 2/3)4/7 (number of reconstructions 2/3) values are stored.
              0: comparison reconstruction 0 and 1:
                 0: reconstruction 0 is much better than reconstruction 1
                 1: reconstruction 0 is better than reconstruction 1
                 3: reconstruction 0 is the same as reconstruction 1
                 4: reconstruction 1 is better than reconstruction 0
                 5: reconstruction 1 is much better than reconstruction 0
              1: image quality of reconstruction 0 determined when comparing with reconstruction 1
                 1: excellent
                 2: good
                 3: diagnostic
                 4: non-diagnostic
              2: image quality of reconstruction 1 determined when comparing with reconstruction 0
                 1: excellent
                 2: good
                 3: diagnostic
                 4: non-diagnostic
              n.a./3: comparison reconstruction 0 and 2:
                 0: reconstruction 0 is much better than reconstruction 2
                 1: reconstruction 0 is better than reconstruction 2
                 3: reconstruction 0 is the same as reconstruction 2
                 4: reconstruction 2 is better than reconstruction 0
                 5: reconstruction 2 is much better than reconstruction 0
              n.a./4: image quality of reconstruction 0 determined when comparing with reconstruction 2
                 1: excellent
                 2: good
                 3: diagnostic
                 4: non-diagnostic
              n.a./5: image quality of reconstruction 2 determined when comparing with reconstruction 0
                 1: excellent
                 2: good
                 3: diagnostic
                 4: non-diagnostic
              3/6 (last): patient number
    """

    def initUI(self):
        # Widgets
        self.addWidget('ExclusivePushButtons', 'image comparison',
                       buttons=['much better', 'better', 'choose..', 'same', 'better', 'much better'],
                       val=2)
        self.addWidget('ComboBox_GROUP', 'image quality', visible=False)
        self.addWidget('PushButton', 'Toggle', button_title='Toggle', toggle=True)
        self.addWidget('TextBox', 'current slice', val='0')
        self.addWidget('TextBox', 'out of slices', val='n.a.')
        self.addWidget('PushButton', 'undo button', button_title='Undo', toggle=False)
        self.addWidget('PushButton', 'Enable image quality rating', button_title='Off', toggle=True)
        self.addWidget('PushButton', 'Enable RMS Normalization', button_title='RMS Normalize', toggle=True)#YCC

        # global variable
        self.nr_recons = 0


        # IO Ports
        self.addInPort('Input_List', 'LIST', obligation=gpi.REQUIRED)
        self.addInPort('previous_analysis_array', 'NPYarray', obligation=gpi.OPTIONAL)

        self.addOutPort('image_for_display', 'NPYarray', ndim=2)
        self.addOutPort('analysis_array', 'NPYarray')

    def validate(self):
        '''This function runs before compute() as a GPI_APPLOOP exec-type.
        Here, widgets (bounds, limits, etc...) can be modified to ensure they
        are correctly validated before the widget values are used in the
        compute() routine -where widgets are buffered and any modifications are
        applied after compute() runs.
        '''
        self.starttime()  # time your code, NODE level log
        import numpy as np
        import random as random

        # examples of node and warn logger levels.
        self.log.node("Running validation()")
        #self.log.warn("this is a bad code area")

        # the path to the gpi.py main script
        #self.log.node("gpi.py path:"+gpi.defines.GPI_CWD)

        # enable image quality widget
        # disable widget if current slice is larger than 0
        if type(self.getData('Input_List')) != type(None):
          if int(self.getVal('current slice')) > 0:
            self.setAttr('Enable image quality rating', visible=False)
          else:
            self.setAttr('Enable image quality rating', visible=True)
        # Depending on 'Enable image quality rating' widget set image quality widget visible or not.
        if self.widgetEvent() == 'Enable image quality rating':
          if self.getVal('Enable image quality rating'):
            self.setAttr('Enable image quality rating', button_title='On')
            self.setAttr('image quality', visible=True)
          else:
            self.setAttr('Enable image quality rating', button_title='Off')
            self.setAttr('image quality', visible=False)
        enable_iq = self.getVal('Enable image quality rating')

        # check for misc event types, these can also be used in compute
        if self.portEvent() == 'Input_List':
          # GETTING PORT INFO
          data = self.getData('Input_List')

          status = 0

          # check whether a patient dataset is loaded
          nr_patients = len(data)
          if nr_patients < 1:
            status = 1
            return status

          # how many recon per dataset? self.nr_recons. Allow 2 and 3 only for now.
          if status == 0:
            self.nr_recons = (np.shape(data[0]))[0]
            if self.nr_recons < 2 and self.nr_recons > 3:
              self.log.warn("Currently only 2 or 3 different recons to be compared are allowed.")
              status = 1
              return status

          # check if all datasets have self.nr_recons recon types
          if status == 0:
            for patient in range(nr_patients):
              self.log.node(np.shape(data[patient]))
              if (np.shape(data[patient]))[0] != self.nr_recons:
                status = 1
                return status

          # count all the slices
          nr_slices = 0
          if status == 0:
            for patient in range(nr_patients):
              nr_slices += (np.shape(data[patient]))[1]
              self.log.node(" nr_slices = "+str(nr_slices))
              if nr_slices==0:
                status = 1
                return status

          # multiply by 2 if self.nr_recons = 3 since we will compare the first data set with both the second and the third
          nr_slices_in_analysis_array = nr_slices
          if self.nr_recons == 3:
            nr_slices *=2
          self.log.node(" nr_slices = "+str(nr_slices))
          self.setAttr('out of slices', val=str(nr_slices))
          self.setAttr('current slice', val='0')

          # set an empty array to analysis_array OutPort
          # size of second dimension is 4 for 2 reconstructions and 7 for 3 reconstructions
          analysis_dim = ( (self.nr_recons-1) * 3 ) + 1
          analysis_array = np.zeros((nr_slices_in_analysis_array+1, analysis_dim))
          # current slice
          analysis_array[-1,0] = 0
          self.setData('analysis_array', analysis_array)

        # for any new InPort event and both InPorts are connected, first check if previous_analysis_array dimensions agree with Input_List
        #    and then assign the previous_analysis_array to the OutPort analysis_array if that has no value yet.
        if self.portEvent() == 'Input_List' or self.portEvent() == 'previous_analysis_array':
          self.log.node('Either Input_List or previous_analysis_array port event.')
          if type(self.getData('Input_List')) != type(None) and type(self.getData('previous_analysis_array')) != type(None):
            self.log.node('Both Input_List and previous_analysis_array ports are connected.')
            previous_analysis_array = self.getData('previous_analysis_array')
            nr_slices = int(self.getVal('out of slices'))
            if self.nr_recons == 3:
              nr_slices /= 2
            if nr_slices + 1 == (np.shape(previous_analysis_array))[0]:
              self.log.node('Both Input_List and previous_analysis_array ports have same number of slices.')
              analysis_array_does_not_exist_yet = True
              if type(self.getData('analysis_array')) != type(None):
                analysis_array = self.getData('analysis_array')
                if analysis_array[-1,0]>1:
                  analysis_array_does_not_exist_yet = False
                  self.log.warn('The previous analysis array was not loaded as there is existing data in the analysis array. Start with a fresh module to continue with old work.')
                  status = 1
                  return status
              if analysis_array_does_not_exist_yet:
                analysis_array = previous_analysis_array
                self.setData('analysis_array', analysis_array)
                self.log.node("previous_analysis_array asigned to analysis array!")
                self.setAttr('current slice', val=str(int(analysis_array[nr_slices,0])))
            else:
              self.log.node('Input_List and previous_analysis_array ports do NOT have same number of slices.')
              status = 1
              return status

        # If Toggle is pressed, mirror the 3 selection widgets. The display will be swapped in compute.
        if self.widgetEvent() == 'Toggle' and type(self.getData('Input_List')) != type(None):
          # image comparison
          self.setAttr('image comparison', val=toggle_compare(self.getVal('image comparison')) )
          # image quality
          if enable_iq:
            image_quality_left_widget_value = (self.getVal('image quality'))['image_quality_left']
            image_quality_right_widget_value = (self.getVal('image quality'))['image_quality_right']
            self.setAttr('image quality', val={'image_quality_left':image_quality_right_widget_value})
            self.setAttr('image quality', val={'image_quality_right':image_quality_left_widget_value})

        if self.widgetEvent() == 'undo button' and type(self.getData('Input_List')) != type(None):
          # GETTING WIDGET INFO
          nr_slices = int(self.getVal('out of slices'))
          current_slice = int(self.getVal('current slice'))

          # remove values in analysis_array of current slice to be undone, and then reduce current slice by one
          # only if current slice is larger than 0
          if current_slice > 0:
            if current_slice == nr_slices:
              self.setAttr('image comparison', visible=True)
              if enable_iq:
                self.setAttr('image quality', visible=True)

            current_slice -= 1

            # randomize the order how the slices will be presented, use a defined seed point to make the order reproducible.
            random.seed(0)
            slice_order = list(range(nr_slices))
            random.shuffle(slice_order)

            analysis_array = self.getData('analysis_array').copy()
            if self.nr_recons == 3 and slice_order[current_slice] >= nr_slices/2:
              actual_slice = slice_order[current_slice] - (nr_slices/2)
              compare_idx = 3
              iq_left = 4
              iq_right = 5
            else:
              actual_slice = slice_order[current_slice]
              compare_idx = 0
              iq_left = 1
              iq_right = 2

            analysis_array[actual_slice,compare_idx] = 0
            analysis_array[actual_slice,iq_left] = 0
            analysis_array[actual_slice,iq_right] = 0
            self.setAttr('current slice', val=str(current_slice))
            analysis_array[-1, 0] = current_slice
            self.setData('analysis_array', analysis_array)

          # reset the 3 widgets
          self.setAttr('image comparison', val=2)
          if enable_iq:
            self.setAttr('image quality', val={'image_quality_left':0})
            self.setAttr('image quality', val={'image_quality_right':0})


        #print self.getEvent() # super set of events
        #print self.portEvent()
        #print self.widgetEvent()

        self.endtime('validate()')  # endtime w/ message
        self.endtime()  # endtime w/o message

        return 0

    def compute(self):
        '''This is where the main algorithm should be implemented.
        '''
        import numpy as np
        import random as random

        self.log.node("Running compute()")

        # GETTING PORT INFO
        data = self.getData('Input_List')

        # GETTING WIDGET INFO
        nr_slices = int(self.getVal('out of slices'))
        current_slice = int(self.getVal('current slice'))
        toggle = self.getVal('Toggle')
        enable_iq = self.getVal('Enable image quality rating')

        nr_patients = len(data)
        self.log.node("current slice is " + str(current_slice) + " out of " + str(nr_slices) + " slices")

        # randomize the order how the slices will be presented, use a defined seed point to make the order reproducible.
        random.seed(0)
        slice_order = list(range(nr_slices))
        random.shuffle(slice_order)

        # once all 3 options have been selected, store results, move to next slice, and reset the 3 widgets
        if self.widgetEvent() == 'image comparison' or self.widgetEvent() == 'image quality':
          self.log.node('image comparison widget value is ' + str(self.getVal('image comparison')))
          if enable_iq:
            image_quality_left_widget_value = (self.getVal('image quality'))['image_quality_left']
            self.log.node('image quality left widget value is ' + str(image_quality_left_widget_value))
            image_quality_right_widget_value = (self.getVal('image quality'))['image_quality_right']
            self.log.node('image quality right widget value is ' + str(image_quality_right_widget_value))
          else:
            image_quality_left_widget_value = -1
            image_quality_right_widget_value = -1
          if self.getVal('image comparison') != 2 and image_quality_left_widget_value != 0 and image_quality_right_widget_value != 0:
            analysis_array = self.getData('analysis_array').copy()
            # enter the patient number in the last entry only when the first slice is being analyzed
            if current_slice == 0:
              temp_slice = 0
              for patient in range(nr_patients):
                nr_slices_per_current_patient = (np.shape(data[patient]))[1]
                for temp_patient_slice in range(nr_slices_per_current_patient):
                  analysis_array[temp_slice,-1] = patient
                  temp_slice += 1

            # use random generator to determine if first image is shown left or right
            random.seed(current_slice)
            left = random.randint(0,1)
            if self.nr_recons == 3 and slice_order[current_slice] >= nr_slices/2:
              actual_slice = slice_order[current_slice] - (nr_slices/2)
              self.log.node('actual slice is: ' + str(actual_slice))
              compare_idx = 3
              if ( left == 1 ) == toggle:
                compare_value = self.getVal('image comparison')
                iq_left = 4
                iq_right = 5
              else:
                compare_value = toggle_compare(self.getVal('image comparison') )
                iq_left = 5
                iq_right = 4
            else:
              actual_slice = slice_order[current_slice]
              self.log.node('actual slice is: ' + str(actual_slice))
              compare_idx = 0
              if ( left == 1 ) == toggle:
                compare_value = self.getVal('image comparison')
                iq_left = 1
                iq_right = 2
              else:
                compare_value = toggle_compare(self.getVal('image comparison') )
                iq_left = 2
                iq_right = 1

            analysis_array[actual_slice,compare_idx] = compare_value
            if enable_iq:
              analysis_array[actual_slice,iq_left] = image_quality_left_widget_value
              analysis_array[actual_slice,iq_right] = image_quality_right_widget_value
            current_slice += 1
            self.setAttr('current slice', val=str(current_slice))
            analysis_array[-1, 0] = current_slice
            self.setData('analysis_array', analysis_array)
            self.setAttr('image comparison', val=2)
            if enable_iq:
              self.setAttr('image quality', val={'image_quality_left':0})
              self.setAttr('image quality', val={'image_quality_right':0})

        # disable widgets once last slice has been analyzed
        if current_slice >= nr_slices:
          self.setAttr('image comparison', visible=False)
          if enable_iq:
            self.setAttr('image quality', visible=False)
          current_image = np.zeros((2,2))
        else:

          # translate slice order to patient number and slice number of that patient
          slice_counter = 0
          current_recon_comparison = -1
          keep_going = True
          for patient in range(nr_patients):
            if slice_order[current_slice] < slice_counter + (np.shape(data[patient]))[1] and keep_going:
              current_patient = patient
              slice_of_current_patient = slice_order[current_slice] - slice_counter
              current_recon_comparison = 0
              keep_going = False
            slice_counter += (np.shape(data[patient]))[1]
          # run a second time for the comparison between data set one and three. Refer to this comparison as current_recon_comparison = 1
          for patient in range(nr_patients):
            if slice_order[current_slice] < slice_counter + (np.shape(data[patient]))[1] and keep_going:
              current_patient = patient
              slice_of_current_patient = slice_order[current_slice] - slice_counter
              current_recon_comparison = 1
              keep_going = False
            slice_counter += (np.shape(data[patient]))[1]

          # use random generator to determine if first image is shown left or right
          random.seed(current_slice)
          left = random.randint(0,1)

          if current_recon_comparison == 0:
            if ( left == 1 ) == toggle:
              left_image_nr = 0
              right_image_nr = 1
            else:
              left_image_nr = 1
              right_image_nr = 0
          elif current_recon_comparison == 1:
            if ( left == 1 ) == toggle:
              left_image_nr = 0
              right_image_nr = 2
            else:
              left_image_nr = 2
              right_image_nr = 0
          else:
            self.log.warn("current_recon_comparison has not been set to either 0 or 1.")

          self.log.node("Current random slice " + str(slice_order[current_slice]) + ", patient number " + str(current_patient) + ", slice number " + str(slice_of_current_patient) + ", recon comparison " + str(current_recon_comparison) + ", left image is " + str(left_image_nr))

          # assign current image for display
          current_dataset = data[current_patient]
          left_image = current_dataset[left_image_nr, slice_of_current_patient,:,:]
          right_image = current_dataset[right_image_nr, slice_of_current_patient,:,:]
          if (self.getVal('Enable RMS Normalization') == True): #YCC
              left_image = left_image / np.sqrt(np.mean(np.square(left_image)))
              right_image = right_image / np.sqrt(np.mean(np.square(right_image)))

          current_image = np.append(left_image, right_image, 1)


        # SETTING PORT INFO
        self.setData('image_for_display', current_image)

        return 0

    def execType(self):
        '''Could be GPI_THREAD, GPI_PROCESS, GPI_APPLOOP'''
        return gpi.GPI_PROCESS



# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Nick Zwart
# Date: 2014mar21

import re
import time
import datetime

import gpi

# get the date and time
def suffix():
    today = datetime.date.today().strftime("%Y.%m.%d")
    ctime = time.strftime("%H.%M.%S")
    suffix = today + "_" + ctime
    return suffix

def new_name(name):
    return name + "_" + suffix()

def unique_name(name, lst):
    # find a suitable name that doesn't already exist
    bname = new_name(name)
    while bname in lst:
        bname = new_name(name)
        time.sleep(1) # wait a second for a unique name
    return bname


class ExternalNode(gpi.NodeAPI):
    """Uses the HDF5 h5py project interface for writing arrays.

    INPUT - numpy array to write

    WIDGETS:
    File Browser - button to launch file browser, and typein widget, to give pathname for output file
    Write Mode - write at any event, or write only with new filename
    Write Now - write right now

    NOTE: This is a simple writer for writing 'dataset' objects from an HDF5 file.
    Since the files can be built in many different ways, this should be considered
    a good starting point for writing code to write your specific format.  More
    general use cases will be added in future releases.
    """

    def initUI(self):

       # Widgets
        self.addWidget(
            'SaveFileBrowser', 'File Browser', button_title='Browse',
            caption='Save File (*.hdf5)', filter='hdf5 (*.hdf5)')
        self.addWidget('StringBox', 'set-name', val='gpidata')
        self.addWidget('PushButton', 'compress (GZIP)', toggle=True)
        self.addWidget('PushButton', 'Write Mode', button_title='Write on New Filename', toggle=True)
        self.addWidget('PushButton', 'Write Now', button_title='Write Right Now', toggle=False)

        # IO Ports
        self.addInPort('in', 'NPYarray')

        # store for later use
        self.URI = gpi.TranslateFileURI

    def validate(self):

        if self.getVal('Write Mode'):
            self.setAttr('Write Mode', button_title="Write on Every Event")
        else:
            self.setAttr('Write Mode', button_title="Write on New Filename")

        fname = self.URI(self.getVal('File Browser'))
        self.setDetailLabel(fname)

        return 0

    def compute(self):

        import h5py

        if self.getVal('Write Mode') or self.getVal('Write Now') or ('File Browser' in self.widgetEvents()):

            fname = self.URI(self.getVal('File Browser'))
            if not fname.endswith('.hdf5'):
                fname += '.hdf5'

            if fname == '.hdf5':
                return 0

            label = self.getVal('set-name')
            if label == '':
                self.log.warn('invalid set-name')
                return 0

            if self.getVal('compress (GZIP)'):
                kwargs = {}
                kwargs['compression'] = 'gzip'
                kwargs['compression_opts'] = 9
                data = self.getData('in')
                f = h5py.File(fname, "a")

                if label in list(f.keys()):
                    label = unique_name(label, list(f.keys()))
                    print("dataset label already exists in file, using \'"+label+"\'")

                dset = f.create_dataset(label, data=data, **kwargs)
                f.close()

            else:
                data = self.getData('in')
                f = h5py.File(fname, "a")

                if label in list(f.keys()):
                    label = unique_name(label, list(f.keys()))
                    print("dataset label already exists in file, using \'"+label+"\'")

                dset = f.create_dataset(label, data=data)
                f.close()

        return(0)



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# author: Jim Pipe
# date: 2012sep

import numpy as np
import gpi


class ExternalNode(gpi.NodeAPI):
    """Clamp, Threshold, or Scale values
       Pass does not operate on the data, but sets the min and max widgets to reflect data min and max
       Clamp takes all values above(below) the max(min) and sets them to the max(min)
       Threshold takes all values above(below) the max(min) and sets them to zero
       Scale "Min" OR "Max" multiplies the data by a constant for the desired min/max
       Scale "Min" AND "Max" multiplies the data by a constant and adds an offset for the desired min & max
       For complex numbers, the operations work on the magnitude, while preserving phase
    """

    def initUI(self):

        # Widgets
        self.addWidget('ExclusivePushButtons', 'operation', buttons=[
                       'Pass', 'Clamp', 'Threshold','Scale'], val=0)
        self.addWidget('ExclusivePushButtons', 'MinMax', buttons=[
                       'Min', 'Max', 'Min&Max'], val=0)
        self.addWidget('DoubleSpinBox', 'Min', val = 0, decimals = 5)
        self.addWidget('DoubleSpinBox', 'Max', val = 0, decimals = 5)

        # IO Ports
        self.addInPort('in', 'NPYarray')
        self.addOutPort('out', 'NPYarray')

    def validate(self):

        if self.getVal('operation') == 0:
          self.setAttr('MinMax',visible=False)
          self.setAttr('Min',visible=True)
          self.setAttr('Max',visible=True)
        elif self.getVal('MinMax') == 0:
          self.setAttr('MinMax',visible=True)
          self.setAttr('Min',visible=True)
          self.setAttr('Max',visible=False)
        elif self.getVal('MinMax') == 1:
          self.setAttr('MinMax',visible=True)
          self.setAttr('Min',visible=False)
          self.setAttr('Max',visible=True)
        else:
          self.setAttr('MinMax',visible=True)
          self.setAttr('Min',visible=True)
          self.setAttr('Max',visible=True)

        if self.getVal('Max') < self.getVal('Min'):
          self.setAttr('Max',val = self.getVal('Min'))

        return(0)

    def compute(self):

        import numpy as np
        import scipy.stats as sp

        data = self.getData('in')
        minmax = self.getVal('MinMax')
        pcts = self.getVal('operation')
        minv = self.getVal('Min')
        maxv = self.getVal('Max')

        if np.iscomplexobj(data):
          dmag = np.abs(data)
          darg = np.angle(data)
          if minmax == 0: #Min
            maxv = np.amax(dmag)
          elif minmax == 1: #Max
            minv = np.amin(dmag)

          if pcts == 0: # Pass
            omag = dmag
            self.setAttr('Min',val=np.amin(omag))
            self.setAttr('Max',val=np.amax(omag))
          elif pcts == 1: # Clamp
            omag = np.clip(dmag,minv,maxv)
          elif pcts == 2: # Thresh
            omag = sp.threshold(dmag,minv,maxv,0)
          elif pcts == 3: # Scale
            if minmax == 0: #Min
              omag = (minv/np.amin(dmag))*dmag
            elif minmax == 1: #Max
              omag = (maxv/np.amax(dmag))*dmag
            elif minmax == 2: #MinMax
              datave = 0.5*(np.amin(dmag)+np.amax(dmag))
              outave = 0.5*(minv+maxv)
              if np.amax(dmag) != np.amin(dmag):
                scale = (maxv-minv)/(np.amax(dmag)-np.amin(dmag))
              else:
                scale = 0
              omag = (scale*(dmag-datave)) + outave

          outdat = omag*(np.cos(darg) + np.sin(darg)*1j)

        else:
          if minmax == 0: #Min
            maxv = np.amax(data)
          elif minmax == 1: #Max
            minv = np.amin(data)

          if pcts == 0: # Pass
            outdat = data
            self.setAttr('Min',val=np.amin(outdat))
            self.setAttr('Max',val=np.amax(outdat))
          elif pcts == 1: # Clamp
            outdat = np.clip(data,minv,maxv)
          elif pcts == 2: # Thresh
            outdat = sp.threshold(data,minv,maxv,0)
          elif pcts == 3: # Scale
            if minmax == 0: #Min
              outdat = (minv/np.amin(data))*data
            elif minmax == 1: #Max
              outdat = (maxv/np.amax(data))*data
            elif minmax == 2: #MinMax
              datave = 0.5*(np.amin(data)+np.amax(data))
              outave = 0.5*(minv+maxv)
              if np.amax(data) != np.amin(data):
                scale = (maxv-minv)/(np.amax(data)-np.amin(data))
              else:
                scale = 0
              outdat = (scale*(data-datave)) + outave

        self.setData('out', outdat)
        return(0)



# Copyright (c) 2014, Dignity Health
# 
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
# 
#     Under either license, the following additional term applies:
# 
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
# 
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
# 
#         This file is part of the GPI core node library.
# 
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
# 
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Jim Pype
# Date: 2014Apr

import numpy as np
import sys
import gpi
import math

class ExternalNode(gpi.NodeAPI):
  
    """Node to get compare 2 numpy arrays
    
    INPUT - Two data sets to be compared - they must be the same size and type

    OUTPUT:
    diff - Difference image (L - R) after optional normalization

    WIDGETS:
    Info: gives RSS Difference and Dot Product after optional normalization
          define D = L - R
          define A* as an array with conjugate element values of the array A
          define A A* as the element-wise product of A and A*
          RSS Difference = sqrt(sum(D D*)), where the sum is over all elements
          Dot Product = sum(L R*), where the sum is over all elements

          When data are normalized, the Dot product is the correlation (assuming 0 mean)

    Normalize: Divide each array A by sqrt(sum(A A*)), where A* is the conjugate of A, and the sum is over all elements
          this effectively makes each array have "unit length"
    """

    def initUI(self):
      self.addWidget('TextBox', 'Info')
      self.addWidget('PushButton', 'Normalize', toggle=True, val=1)
      self.addInPort('inLeft', 'NPYarray', dtype=[np.complex64, np.complex128, np.float32, np.float64])
      self.addInPort('inRight','NPYarray', dtype=[np.complex64, np.complex128, np.float32, np.float64])
      self.addOutPort('diff', 'NPYarray')
        
    def validate(self):
      data1 = self.getData('inLeft')
      data2 = self.getData('inRight')
        
      data1_shape = list(data1.shape)
      data2_shape = list(data2.shape)
      if data1_shape != data2_shape:
        self.log.warn("Two data sets must be the same size")
        return(1)
      if np.iscomplexobj(data1) != np.iscomplexobj(data2):
        self.log.warn("Two data sets must both be complex or both be real")
        return(1)
        
      return(0)

    def compute(self):
      
        import sys
        import numpy as np
        import scipy as sp
        
        data1 = self.getData('inLeft')
        data2 = self.getData('inRight')
        normalize = self.getVal('Normalize')

        if np.iscomplexobj(data1): # This means they both are complex
          if normalize:
            # Dividing arrays by their root-sum-square values makes them "unit vectors" in n-dim space
            rss1 = np.sqrt(np.sum(data1*np.conj(data1)))
            rss2 = np.sqrt(np.sum(data2*np.conj(data2)))
            if rss1 > 0:
              data1 = data1/rss1
            if rss2 > 0:
              data2 = data2/rss2
          datdiff = data1-data2
          rssd = np.sqrt(np.sum(datdiff*np.conj(datdiff)))
          dot = np.sum(data1*np.conj(data2))
      
        else: # Data are real-valued
          if normalize:
            rss1 = np.sqrt(np.sum(data1*data1))
            rss2 = np.sqrt(np.sum(data2*data2))
            if rss1 > 0:
              data1 = data1/rss1
            if rss2 > 0:
              data2 = data2/rss2
          datdiff = data1-data2
          rssd = np.sqrt(np.sum(datdiff*datdiff))
          dot = np.sum(data1*data2)

        self.setData('diff',datdiff)
        info = "RSS Difference = "+str(rssd)+'\n\n'+"Dot Product ="+str(dot)
        self.setAttr('Info', val=info)
    
        return(0)



# Copyright (c) 2014, Dignity Health
#
#     The GPI core node library is licensed under
# either the BSD 3-clause or the LGPL v. 3.
#
#     Under either license, the following additional term applies:
#
#         NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL
# PURPOSES AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
# SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
# PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
# USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT LIMITED
# TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR MAKES NO
# WARRANTY AND HAS NOR LIABILITY ARISING FROM ANY USE OF THE SOFTWARE IN ANY
# HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#     If you elect to license the GPI core node library under the LGPL the
# following applies:
#
#         This file is part of the GPI core node library.
#
#         The GPI core node library is free software: you can redistribute it
# and/or modify it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version. GPI core node library is distributed
# in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.
#
#         You should have received a copy of the GNU Lesser General Public
# License along with the GPI core node library. If not, see
# <http://www.gnu.org/licenses/>.


# Author: Jim Pipe
# Date: 2013Aug

import gpi


class ExternalNode(gpi.NodeAPI):

    """Module to generate the gradient waveforms for a desired k-space waveform.  Uses the core file bnispiralgen.c
    If using this module to generate k-space coordinates for collected data, it is crucial that the code be identical
    with that used in the methods code (hence we keep bnispiralgen.c as a standalone file)

    INPUT: dictionary from (e.g.) ReadPhilips with all of the parameters used to collect data.  If input data are present,
           parameters cannot be changed.  If not, parameters may be changed freely.

    OUTPUTS:
    crds_out - output coordinates: the last dimension is either 2 (kx/ky) or 3 (kx/ky/kz) for 2D and 3D trajectories.  See
               discussion of "true resolution" below for more details about the scaling.  Dwell time specified by AD dwell time widget.
    grd_out - gradient waveforms used to produce crds_out.  Fixed to dwell time of 6.4 us.

    WIDGETS: most widgets self-explanatory, here are a few clarifications:
    stype - ARCH is standard 2D spiral
            CYL DST is a cylindrical distributed spiral
            SPH DST is a cylindrical distributed spiral
            FLORET is FLORET...
    # of spiral arms - allows floating point entry
                       for ARCH this is rounded to integer
                       for CYL, SPH DST and FLORET, which share arms across ky and kz, this is the effective arms per plane
                       and can be a floating point number (giving greater flexibility in choosing the total number of arms vs.
                       readout duration)
    usamp st (0-1) - the relative value of kr at which undersampling begins (0 at the center, 1 at the edge of collected k-space)
                     the samples are collected at the nyquist limit (1/FOV X-Y) prior to that
    usamp end (0-1) - the relative value of kr at which undersampling ends
                     the samples are collected at the R times the nyquist limit (1/FOV X-Y) prior to that
    max underample - R, i.e. the undersampling relative to (1/FOV X-Y) after kr exceeds usamp end
    ustype - determines the progression of radial undersampling from usamp st to usamp end as linear, quadratic, or hanning
    Max G Freq - limits the maximum frequency of the gradient waveform during the spiral readout.  If set to 0, there is no limit
    T2 Match - allows one to exponentially change the radial undersampling with time to match T2* decay, for optimal, T2*-matched, SNR
    Sloppy Sp. Per - changes the characteristic of the sloppy spiral, if desired (see Miki Lustig's perturbed spiral paper).  0 turns off.
    gtype - amount of gradient to output
        SPIRAL ONLY - the gradient waveform during the readout only
        G -> 0 - Same as above, but with gradient rampdown
        M0 -> 0 - Same as above, but with k-space rewinder
        M1 -> 0 - Same as above, but with gradient first moment compensation
    spinout - controls spiral in, out, etc.
        OUT - generate spiral-out waveform
        IN  - generate spiral-in waveform
        INOUT 180R - generate single waveform with spiral-in followed by spiral-out on a trajectory rotated by 180 degrees
        INOUT SAME - generate single waveform with spiral-in followed by spiral-out on the same trajectory

    ****************************************
    *** The concept of "true resolution" ***
    k-space data are typically normalized, for the gridding process, to values between -0.5 and 0.5
    For spiral data (which measure circular k-space) to have the same resolution as Cartesian (square k-space)
    they must measure a diameter of 2/sqrt(pi) ~ 1.13 larger than conventional k-space limits (so that the area of the circle
    equals the area of the square).  k-space coordinates, therefore, are multiplied by 0.8, so that there range of
    0.8*2/sqrt(pi) = 0.903, or -0.451 to +0.451, fits within the gridded space.  The resulting image, with no further
    zero-padding, will have pixels that are 0.8 times smaller than the requested resolution, with a matrix 25% larger in each
    dimension.  This is a semi-complicated way of making sure that this works routinely, and is referred to by the authors
    as "true resolution".

    A similar logic is used for measuring data in a sphere, but the diameter must be the cubed root of (6/pi), or 1.24, the
    side of an equivalent cube for equal volumes.  k-space coordinates then span 0.8*(6/pi)^(1/3) = 0.993 of the unit width of
    gridded k-space.
    *****************************************
    """

    def execType(self):
        return gpi.GPI_PROCESS

    def initUI(self):

        # Widgets
        self.addWidget('PushButton', 'compute', toggle=True)
        self.addWidget('TextBox', 'Info:')

        self.addWidget('DoubleSpinBox', 'AD dwell time (us)',
                       val=1.0, min=0.1, decimals=6)
        self.addWidget('DoubleSpinBox', 'x delay (us)',
                       val=0.0, min=-100.0, max=100.0)
        self.addWidget('DoubleSpinBox', 'y delay (us)',
                       val=0.0, min=-100.0, max=100.0)
        self.addWidget('DoubleSpinBox', 'z delay (us)',
                       val=0.0, min=-100.0, max=100.0)

        self.addWidget('DoubleSpinBox', 'MaxSlw (mT/m/ms)',
                       val=150.0, min=0.01, decimals=6)
        self.addWidget('DoubleSpinBox', 'MaxGrd (mT/m)',
                       val=40.0, min=0.01, decimals=6)
        self.addWidget('DoubleSpinBox', 'Gam (kHz/mT)',
                       val=42.577, min=0.01, decimals=6)

        self.addWidget('DoubleSpinBox', 'Fov X-Y (cm)',
                       val=24.0, min=0.1, decimals=6)
        self.addWidget('DoubleSpinBox', 'Fov Z (cm)',
                       val=24.0, min=0.1, decimals=6)
        self.addWidget('DoubleSpinBox', 'ImgRes X-Y (cm)',
                       val=0.2, min=0.01, singlestep=0.01, decimals=6)
        self.addWidget('DoubleSpinBox', 'ImgRes Z (cm)', val=0.2,
                       min=0.01, singlestep=0.01, decimals=6)

        self.addWidget('ExclusivePushButtons', 'stype',
                       buttons=['ARCH', 'CYL DST', 'SPH DST', 'FLORET'], val=0)
        self.addWidget('DoubleSpinBox', 'Taper', val=0.0, min=0.0, max = 1.0)
        self.addWidget('DoubleSpinBox', '# of Spiral Arms', val=16.0, min=0.0)

        self.addWidget('SpinBox', '# of Hubs', val=3, min=1, max=3)
        self.addWidget('DoubleSpinBox', 'Alpha0', val=45.0, min=0.0, max=90.0)
        self.addWidget('PushButton', 'FLORET Rebin', toggle=True)

        self.addWidget('DoubleSpinBox', 'usamp st (0 - 1)',
                       val=0.0, min=0.0, max=1.0, singlestep=0.01)
        self.addWidget('DoubleSpinBox', 'usamp end (0 - 1)',
                       val=1.0, min=0.0, max=1.0, singlestep=0.01)
        self.addWidget('DoubleSpinBox', 'max undersample',
                       val=1.0, min=0.0, max=100.0)
        self.addWidget('ExclusivePushButtons', 'ustype',
                       buttons=['LINEAR', 'QUAD', 'HANNING'], val=0)

        self.addWidget('DoubleSpinBox', 'Max G Freq (kHz)', val=0.0, min=0.0)
        self.addWidget('DoubleSpinBox', 'T2 Match (ms)', val=0.0, min=0.0)

        self.addWidget('DoubleSpinBox',
                       'Sloppy Sp. Per (0:off)', val=0.0, min=0.0)

        self.addWidget('ExclusivePushButtons', 'gtype',
                       buttons=['SPIRAL ONLY', 'G -> 0', 'M0 -> 0', 'M1 -> 0', 'Fast Spoil'], val=2)

        self.addWidget('ExclusivePushButtons', 'spinout',
                       buttons=['OUT', 'IN', 'INOUT 180R', 'INOUT SAME', 'INOUT ROT2', 'INOUT SAME2'], val=0)
        self.addWidget('SpinBox', 'Num Calibration Points', val=0, min=0, max=512)

        # IO Ports
        self.addInPort('params_in', 'DICT', obligation=gpi.OPTIONAL)
        self.addInPort('arm_0', 'NPYarray', obligation=gpi.OPTIONAL)
        self.addOutPort('crds_out', 'NPYarray')
        self.addOutPort('grd_out', 'NPYarray')
        self.addOutPort('kmag_trace', 'NPYarray')

    def validate(self):

        inparam = self.getData('params_in')

        if (inparam is not None):

            if 'headerType' in inparam:

                # check if the header is for spiral
                if inparam['headerType'] != 'BNIspiral':
                    self.log.warn("wrong header type")
                    return 1

            else:
                # if there is no header type, then its also the wrong type
                self.log.warn("wrong header type")
                return 1

            self.setAttr('AD dwell time (us)', val=1000. *
                         float(inparam['spDWELL'][0]))

            self.setAttr('MaxSlw (mT/m/ms)',
                         val=float(inparam['spSLEWMAX'][0]))
            self.setAttr('MaxGrd (mT/m)',
                         val=float(inparam['spGMAX'][0]))
            self.setAttr('Gam (kHz/mT)',
                         val=float(inparam['spGAMMA'][0]))

            self.setAttr('Fov X-Y (cm)',
                         val=100. * float(inparam['spFOVXY'][0]))
            self.setAttr('Fov Z (cm)',
                         val=100. * float(inparam['spFOVZ'][0]))
            self.setAttr('ImgRes X-Y (cm)',    val=100. *
                         float(inparam['spRESXY'][0]))
            self.setAttr('ImgRes Z (cm)',
                         val=100. * float(inparam['spRESZ'][0]))

            self.setAttr('stype',
                         val=int(float(inparam['spSTYPE'][0])))
            self.setAttr('# of Spiral Arms',
                         val=float(inparam['spARMS'][0]))

            self.setAttr('usamp st (0 - 1)',
                         val=float(inparam['spUS0'][0]))
            self.setAttr('usamp end (0 - 1)',
                         val=float(inparam['spUS1'][0]))
            self.setAttr('max undersample',
                         val=float(inparam['spUSR'][0]))
            self.setAttr('ustype',
                         val=int(float(inparam['spUSTYPE'][0])))
            self.setAttr('gtype',
                         val=int(float(inparam['spGTYPE'][0])))
            if 'spMGFRQ' in inparam:
                self.setAttr('Max G Freq (kHz)',
                             val=(float(inparam['spMGFRQ'][0])))

            if 'spT2MATCH' in inparam:
                self.setAttr('T2 Match (ms)',
                             val=(float(inparam['spT2MATCH'][0])))

            spinout = int(inparam['spINOUT_ON'][0])
            spinout += int(inparam['spINOUT_OPT'][0])
            self.setAttr('spinout',      val=spinout)

            if spinout > 1: # in-out
                self.setAttr('Num Calibration Points', visible=True)
                if 'spEXTRA_GRAD_PNT' in inparam:
                    self.setAttr('Num Calibration Points', val=int(0.5*float(inparam['spEXTRA_GRAD_PNT'][0])))
            else:
                self.setAttr('Num Calibration Points', visible=False)

            # FLORET params
            if 'spHUBS' in inparam:
                hubs = int(float(inparam['spHUBS'][0]))
            else:
                hubs = self.getVal('# of Hubs')
            if 'spALPHA0' in inparam:
                alpha0 = float(inparam['spALPHA0'][0])
            else:
                alpha0 = self.getVal('Alpha0')
            if 'spFLORETbin' in inparam:
                rebin = float(inparam['spFLORETbin'][0])
            else:
                rebin = self.getVal('FLORET Rebin')
            if self.getVal('stype') == 3:
                self.setAttr('# of Hubs', visible=True, val=hubs)
                self.setAttr('Alpha0', visible=True, val=alpha0)
                self.setAttr('FLORET Rebin', visible=True, val=rebin)
            else:
                self.setAttr('# of Hubs', visible=False)
                self.setAttr('Alpha0', visible=False)
                self.setAttr('FLORET Rebin', visible=False, val=0)

        self.setAttr('Taper', visible=(self.getVal('stype') == 1)) # CDST

        if self.getVal('stype') == 3:
            rebin = self.getVal('FLORET Rebin')
            self.setAttr('# of Hubs', visible=True)
            self.setAttr('Alpha0', visible=True)
            self.setAttr('FLORET Rebin', visible=True, val=rebin)
        else:
            self.setAttr('# of Hubs', visible=False)
            self.setAttr('Alpha0', visible=False)
            self.setAttr('FLORET Rebin', visible=False, val=0)

    def compute(self):

        import numpy as np

        arm_0 = self.getData('arm_0')
        # arm_0 = np.insert(np.diff(arm_0,0),0,0)

        # convert units to ms, kHz, m, mT
        dwell = 0.001 * self.getVal('AD dwell time (us)')
        xdely = 0.001 * self.getVal('x delay (us)')
        ydely = 0.001 * self.getVal('y delay (us)')
        zdely = 0.001 * self.getVal('z delay (us)')

        mslew = self.getVal('MaxSlw (mT/m/ms)')
        mgrad = self.getVal('MaxGrd (mT/m)')
        gamma = self.getVal('Gam (kHz/mT)')

        fovxy = 0.01 * self.getVal('Fov X-Y (cm)')
        fovz = 0.01 * self.getVal('Fov Z (cm)')
        resxy = 0.01 * self.getVal('ImgRes X-Y (cm)')
        resz = 0.01 * self.getVal('ImgRes Z (cm)')

        stype = self.getVal('stype')
        narms = self.getVal('# of Spiral Arms')
        taper = self.getVal('Taper')

        try:
            hubs = self.getVal('# of Hubs')
        except:
            hubs = 3

        try:
            alpha0 = np.pi * self.getVal('Alpha0') / 180.0
        except:
            alpha0 = 0.25 * np.pi

        try:
            rebin = self.getVal('FLORET Rebin')
            if rebin:
                res_sphere = np.power(np.pi / 6, (1./3.))
                nspiral = round(narms * alpha0 * (fovz / (resz * res_sphere)))
                binfact = int(round(nspiral / 34.0))
                nspiral = int(34 * binfact)
                narms = nspiral / (alpha0 * (fovz / (resz * res_sphere)))
                self.setAttr('# of Spiral Arms', val=narms)
        except:
            rebin = 0

        us_0 = self.getVal('usamp st (0 - 1)')
        us_1 = self.getVal('usamp end (0 - 1)')
        us_r = self.getVal('max undersample')
        utype = self.getVal('ustype')

        mgfrq = self.getVal('Max G Freq (kHz)')
        t2mch = self.getVal('T2 Match (ms)')

        slper = self.getVal('Sloppy Sp. Per (0:off)')
        gtype = self.getVal('gtype')
        spinout = self.getVal('spinout')
        if spinout > 1:
            numCalPnts = self.getVal('Num Calibration Points')
        else:
            numCalPnts = 0


        # read num of data points from parameter txt file
        numREADPTS = 0
        inparam = self.getData('params_in')
        if (inparam is not None):
            if 'spREADPTS' in inparam:
                numREADPTS = int(float(inparam['spREADPTS'][0]))

        if self.getVal('compute'):

            # import in thread to save namespace
            import gpi_core.spiral.spiral as sp
            if stype == 0:
                zdely = 0.0

            # determine k-space coordinates based on input of first arm
            # TODO: doesn't handle spinout or FLORET rebinning (RKR)
            if arm_0 is not None:
                npts = arm_0.shape[0]
                crds_out = np.zeros((hubs,narms,npts,3))

                # goldangle = np.pi*(3-np.sqrt(5))
                # goldangle = 137.508*np.pi/180.
                goldangle = 137.508*3.141592653589793/180.
                if stype == 0: # arch
                    for arm in range(np.int(narms)):
                        beta = -arm*2*np.pi/narms
                        crds_out[0,arm,:,0] = np.cos(beta)*arm_0[:,0] - np.sin(beta)*arm_0[:,1]
                        crds_out[0,arm,:,1] = np.cos(beta)*arm_0[:,1] + np.sin(beta)*arm_0[:,0]

                elif stype == 1: # cylindrical distributed spirals
                    for arm in range(np.int(narms)):
                        beta = -arm*goldangle
                        crds_out[0,arm,:,0] = np.cos(beta)*arm_0[:,0] - np.sin(beta)*arm_0[:,1]
                        crds_out[0,arm,:,1] = np.cos(beta)*arm_0[:,1] + np.sin(beta)*arm_0[:,0]
                        crds_out[0,arm,:,2] = -(2*arm/narms - 1)*arm_0[:,2]

                elif stype == 2: # spherical distributed spirals
                    for arm in range(np.int(narms)):
                        beta = -arm*goldangle
                        crds_out[0,arm,:,0] = np.cos(beta)*arm_0[:,0] - np.sin(beta)*arm_0[:,1]
                        crds_out[0,arm,:,1] = np.cos(beta)*arm_0[:,1] + np.sin(beta)*arm_0[:,0]
                        crds_out[0,arm,:,2] = -(2*arm/narms - 1)*arm_0[:,2]

                elif stype == 3: # FLORET
                    for arm in range(np.int(narms)):
                        beta  = -arm*goldangle
                        alpha = -alpha0 + arm/narms*alpha0*2.
                        crds_out[0,arm,:,0] = np.cos(alpha)*(np.cos(beta)*arm_0[:,0] - np.sin(beta)*arm_0[:,1])/np.cos(-alpha0)
                        crds_out[0,arm,:,1] = np.cos(alpha)*(np.cos(beta)*arm_0[:,1] + np.sin(beta)*arm_0[:,0])/np.cos(-alpha0) # OK
                        crds_out[0,arm,:,2] = np.sin(alpha)*arm_0[:,2]/np.sin(-alpha0)

                    if not (xdely == ydely == zdely):
                        self.log.warn("FLORET trajectory calculation based on input of first arm can not account for anisotropic gradient delays!")

                    if hubs > 1:
                        crds_out[1,:,:,0] = crds_out[0,:,:,1] # OK
                        crds_out[1,:,:,1] = -crds_out[0,:,:,2]
                        crds_out[1,:,:,2] = -crds_out[0,:,:,0]
                    if hubs > 2:
                        crds_out[2,:,:,0] = crds_out[0,:,:,2]
                        crds_out[2,:,:,1] = crds_out[0,:,:,1] # OK
                        crds_out[2,:,:,2] = -crds_out[0,:,:,0]

                else:
                    pass

                self.log.warn("True gradient waveform output not (yet) implemented for calculation based on initial spiral arm!")
                grd_out = np.diff(crds_out, axis=-2)
            else:
                grd_out, crds_out = sp.coords(
                    dwell, xdely, ydely, zdely, mslew, mgrad, gamma, fovxy, fovz,
                    resxy, resz, stype, narms, taper, hubs, alpha0, rebin, us_0, us_1, us_r,
                    utype, mgfrq, t2mch, slper, gtype, spinout, numCalPnts)

            # Report Back to User
            nsamp = np.array(crds_out.shape)[2]
            # match data points from parameter file or real data
            if (numREADPTS > 0) and (nsamp > numREADPTS):
                ndim = np.array(crds_out.shape)[-1]
                crds_out = crds_out[..., 0:numREADPTS, 0:ndim]
                nsamp = numREADPTS

            ntrs = np.array(crds_out.shape)[0] * np.array(crds_out.shape)[1]
            # For Philips grad, assume PHGRAST = 0.0064 ms
            tgrad = 0.0064 * np.array(grd_out.shape)[1]
            smp = "Samples: " + str(nsamp) + "\n"
            tau = "Tau (ms): " + str(dwell * nsamp) + "\n"
            tgr = "TGrad (ms): " + str(tgrad) + "\n"
            trs = "Total interleaves: " + str(ntrs)
            info = smp + tau + tgr + trs
            self.setAttr('Info:', val=info)

            # if ARCH set 2d spiral cords out DHW
            if stype == 0:  # ARCH
                crds_out = crds_out[0, ..., 0:2]
                grd_out = grd_out[..., 0:2]
            elif stype != 3:  # FLORET
                crds_out = crds_out[0, ...]

            self.setData('crds_out', crds_out)
            self.setData('grd_out', grd_out)
            self.setData('kmag_trace', np.sqrt(np.sum(crds_out[0,...]**2, axis=-1)))

        return(0)
